"""Defines objects used in mesing operations."""

from __future__ import annotations

from itertools import count
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt  # type: ignore
import numpy as np
from descartes.patch import PolygonPatch  # type: ignore
from matplotlib.patches import Polygon
from shapely.geometry import Polygon as shapely_Polygon  # type: ignore

from osmg.core import common

if TYPE_CHECKING:
    from osmg.core.common import numpy_array


class Vertex:
    """
    2D Vertex.

    It knows all the edges connected to it.
    It knows all the halfedges leaving from it.
    Each instance has an automatically generated unique id.

    Attributes:
    ----------
        coordinates: Coordinates of the vertex.
        edges: List of edges connected to the vertex.
        halfedges: List of halfedges leaving from the vertex.
        uid: Unique identifier of the vertex.

    Example:
        >>> from osmg.mesh import Vertex
        >>> v = Vertex((0.0, 0.0))

    """

    _ids = count(0)

    def __init__(self, coordinates: tuple[float, float]) -> None:
        """
        Initialize a new instance of the `Vertex` class.

        Arguments:
            coordinates: Coordinates of the vertex.

        """
        self.coordinates = coordinates
        self.edges: list[Edge] = []
        self.halfedges: list[Halfedge] = []
        self.uid: int = next(self._ids)

    def __eq__(self, other: object) -> bool:
        """
        Check for equality based on the uid of the vertex.

        Arguments:
            other: Other vertex to compare with.

        Returns:
        -------
            bool: `True` if the two vertices are equal, `False`
                  otherwise.

        Example:
            >>> from osmg.mesh import Vertex
            >>> v1 = Vertex((0, 0))
            >>> v2 = Vertex((1, 1))
            >>> v3 = Vertex((0, 0))
            >>> v1 == v2
            False
            >>> v1 == v3
            False
            >>> v1 == v1
            True

        """
        assert isinstance(other, Vertex)
        return bool(self.uid == other.uid)

    def __hash__(self) -> int:
        """
        Obtain a hash value for the vertex based on its uid.

        Returns:
        -------
            int: Hash value of the vertex.

        Example:
            >>> from osmg.mesh import Vertex
            >>> v1 = Vertex((0, 0))
            >>> hash(v1)
            0
            >>> v2 = Vertex((1, 1))
            >>> hash(v2)
            1
        """
        return hash(self.uid)

    def __repr__(self) -> str:
        """
        Obtain a string representation of the vertex.

        Returns:
        -------
            String representation of the vertex.

        """
        return f'(V{self.uid} @ {self.coordinates}) '


class Edge:
    """
    2D oriented Edge.

    Connected to two vertices `v_i` and `v_j`.  Has
    two halfedges, `h_i` and `h_j`.  Each instance has an
    automatically generated unique id.

    """

    _ids = count(0)

    def __init__(self, v_i: Vertex, v_j: Vertex) -> None:
        """
        Initialize a new edge with the given vertices.

        If the vertices don't already have an edge connecting them,
        this edge is added to their list of edges.

        """
        self.v_i = v_i
        self.v_j = v_j
        self.uid = next(self._ids)
        self.h_i: Halfedge | None = None
        self.h_j: Halfedge | None = None
        if self not in self.v_i.edges:
            self.v_i.edges.append(self)
        if self not in self.v_j.edges:
            self.v_j.edges.append(self)

    def __repr__(self) -> str:
        """
        Obtain a string representation of this edge.

        Returns:
          A string representation of this edge in the form
          `(E{self.uid} @ V{self.v_i.uid}, V{self.v_j.uid})`

        """
        return f'(E{self.uid} @ V{self.v_i.uid}, V{self.v_j.uid}) '

    def define_halfedge(self, vertex: Vertex) -> Halfedge:
        """
        Define a halfedge.

        For the current edge instance and given one of its vertices,
        we want the halfedge that points to the direction
        away from the given vertex.
        We create it if it does not exist.

        Returns:
          The halfedge.

        Raises:
          ValueError: If the halfedge is already defined.
        """
        if vertex == self.v_i:
            if not self.h_i:
                halfedge = Halfedge(self.v_i, self)
                self.h_i = halfedge
            else:
                msg = 'Halfedge h_i already defined'
                raise ValueError(msg)
        elif vertex == self.v_j:
            if not self.h_j:
                halfedge = Halfedge(self.v_j, self)
                self.h_j = halfedge
            else:
                msg = 'Halfedge h_j already defined'
                raise ValueError(msg)
        else:
            msg = 'The edge is not connected to the given vertex.'
            raise ValueError(msg)
        return halfedge

    def other_vertex(self, vertex: Vertex) -> Vertex:
        """
        Obtain the vertex of this edge that is not the given vertex.

        If the given vertex is not connected to this edge, a ValueError
        is raised.

        Example:
            >>> from osmg.mesh import Vertex
            >>> v1 = Vertex((0, 0))
            >>> v2 = Vertex((1, 0))
            >>> v3 = Vertex((2, 0))
            >>> e = Edge(v2, v3)
            >>> e.other_vertex(v2).coordinates
            (2, 0)
            >>> e.other_vertex(v3).coordinates
            (1, 0)
            >>> e.other_vertex(v1)
            Traceback (most recent call last):
                ...
            ValueError: The edge is not connected to the given vertex

        Returns:
          The other vertex.

        Raises:
          ValueError: If the edge is not connected to the given vertex.
        """
        if self.v_i == vertex:
            v_other = self.v_j
        elif self.v_j == vertex:
            v_other = self.v_i
        else:
            msg = 'The edge is not connected to the given vertex'
            raise ValueError(msg)
        return v_other

    def overlaps_or_crosses(self, other: Edge) -> bool:
        """
        Obtain True if this edge overlaps or crosses another edge.

        Edges are allowed to share one vertex (returns False), but not
        both (returns True).

        Arguments:
            other : Edge
                The other edge to check for overlap or cross with this
                edge.

        Example:
            >>> from osmg.mesh import Vertex, Edge
            >>> v1 = Vertex((0, 0))
            >>> v2 = Vertex((0, 1))
            >>> v3 = Vertex((1, 1))
            >>> v4 = Vertex((1, 0))
            >>> e1 = Edge(v1, v2)
            >>> e2 = Edge(v3, v4)
            >>> e1.overlaps_or_crosses(e2)
            False
            >>> e2.overlaps_or_crosses(e1)
            False
            >>> e3 = Edge(v1, v3)
            >>> e4 = Edge(v2, v4)
            >>> e3.overlaps_or_crosses(e4)
            True
            >>> e4.overlaps_or_crosses(e3)
            True
            >>> e5 = Edge(v1, v4)
            >>> e6 = Edge(v2, v3)
            >>> e5.overlaps_or_crosses(e6)
            False
            >>> e6.overlaps_or_crosses(e5)
            False

        Returns:
            True if this edge overlaps or crosses the other edge,
            False otherwise.

        """
        # location of this edge
        vec_ra: numpy_array = np.array(self.v_i.coordinates)
        # direction of this edge
        vec_da: numpy_array = np.array(self.v_j.coordinates) - np.array(
            self.v_i.coordinates
        )
        # location of other edge
        vec_rb: numpy_array = np.array(other.v_i.coordinates)
        # direction of other edge
        vec_db: numpy_array = np.array(other.v_j.coordinates) - np.array(
            other.v_i.coordinates
        )
        # verify that the edges have nonzero length
        assert not np.isclose(vec_da @ vec_da, 0.00)
        assert not np.isclose(vec_db @ vec_db, 0.00)

        mat_a: numpy_array = np.column_stack((vec_da, -vec_db))
        mat_b = vec_rb - vec_ra
        determinant = np.linalg.det(mat_a)

        if np.isclose(determinant, 0.00):
            # there are infinite solutions
            # or there are no solutions
            # i.e., the edges are parallel.
            # If they are parallel but nor colinear, then they don't
            # overlap and the method should return False
            # If they are colinear, then they might overlap. If they
            # do, the method should return True, otherwise False.

            # first check if they are parallel but not colinear
            # project start of other vertex onto line of this vertex
            vec_rb_diff = vec_rb - vec_ra
            vec_proj_pt = (vec_rb_diff @ vec_da) / (
                vec_da @ vec_da
            ) * vec_da + vec_ra
            vec_dist = vec_rb - vec_proj_pt
            distance = np.sqrt(vec_dist @ vec_dist)

            if not np.isclose(distance, 0.00):
                # The edges are parallel but not collinear, so they
                # can't be intersecting.
                return False

            # If the previous statement was not true, we will arrive
            # here. The edges are colinear. Depending on their
            # relative position on their common line, they might share
            # no common points, one common point, or an entire
            # segment.
            # To solve this, we define ta to be a scalar that
            # determines a point on vertex i by evaluating: vec_ra +
            # ta * vec_da.
            # ta = 0 ==> on vertex_i, ta = 1 ==> on vertex j, of this
            # edge.
            # Similarly, there exists a tb that can be used to
            # identify a point on vertex j
            # But insdtead, we determine the location of vertex i and
            # j of the other edge in terms of ta. That is:
            # ta = c_i ==> vertex i of other edge, ta = c_j ==> vertex
            # j of other edge.
            # We can then determine which of the three cases we are
            # in, based on the values of c_i and c_j.
            c_i = (vec_da @ (vec_rb - vec_ra)) / (vec_da @ vec_da)
            c_j = (vec_da @ ((vec_rb + vec_db) - vec_ra)) / (vec_da @ vec_da)
            # either they should be both < 0 (which means that the
            # other edge is "before" this edge), or they should be
            # both > 1.00 (which means that the other edge is "after"
            # this edge). Any other case corresponds to an overlap.

            # each of c_i, c_j can either be {<0.00, ==0, 00<1.00, ==1, >1.0}
            # in each case the answer will depend on what the other one is.
            # note: we need to account for floating-point precision
            # when making comparisons.
            epsilon = common.EPSILON
            if (
                (c_i < 0.00 - epsilon and np.isclose(c_j, 0.00))  # noqa: PLR0916
                or (c_i > 1.00 + epsilon and np.isclose(c_j, 1.00))
                or (np.isclose(c_i, 1.00) and c_j > 1.00 + epsilon)
                or (np.isclose(c_i, 0.00) and c_j < 0.00 - epsilon)
            ):
                # they share one vertex without overlap
                return False
            return not (
                (c_i < 0.00 - epsilon and c_j < 0.00 - epsilon)
                or (c_i > 1.00 + epsilon and c_j > 1.00 + epsilon)
            )

        # Otherwise they are not parallel.
        # there is at least one solution
        sol = np.linalg.solve(mat_a, mat_b)
        # if both constants are between 0 and 1
        # the edges overlap within their length
        # otherwise, their extensions overlap, which
        # is not an issue.
        return bool(0.0 < sol[0] < 1.0 and 0.0 < sol[1] < 1.0)


class Halfedge:
    """
    Halfedge object.

    Every edge has two halfedges. A halfedge has a direction, pointing
    from one of the corresponding edge's vertices to the other.  The
    `vertex` attribute corresponds to the edge's vertex that the
    halfedge originates from.  Halfedges have a `next` attribute that
    points to the next halfedge, forming closed loops, or sequences,
    which is the purpose of this module.

    """

    _ids = count(0)

    def __init__(self, vertex: Vertex, edge: Edge) -> None:
        """
        Initialize the halfedge object.

        Arguments:
            vertex: The vertex that the halfedge originates from.
            edge: The edge that the halfedge is a part of.

        """
        self.vertex = vertex
        self.edge = edge
        self.uid: int = next(self._ids)
        self.nxt: Halfedge | None = None

    def __repr__(self) -> str:
        """
        Obtain a string representation of the halfedge.

        Returns:
          A string representation of the halfedge, in the form
          `(H0 from E0 to E0 next H1)`

        """
        if self.nxt:
            out = (
                f'(H{self.uid} from E{self.edge.uid}'
                f' to E{self.nxt.edge.uid} next H{self.nxt.uid})'
            )
        else:
            out = f'(H{self.uid}'
        return out

    def __lt__(self, other: object) -> bool:
        """
        Comparison function used for sorting. Compares the halfedge ids.

        Returns:
          Whether the other object is less than the current.
        """
        assert isinstance(other, Halfedge)
        return bool(self.uid < other.uid)

    def direction(self) -> float:
        """
        Calculate the angular direction of the halfedge.

        Calculates the angular direction of the halfedge
        using the arctan2 function (in radians).

        Example:
            >>> from osmg.mesh import Vertex, Edge, Halfedge
            >>> v1 = Vertex((0.0, 0.0))
            >>> v2 = Vertex((2.0, 2.0))
            >>> edge = Edge(v1, v2)
            >>> halfedge1 = Halfedge(v1, edge)
            >>> halfedge2 = Halfedge(v2, edge)
            >>> halfedge1.direction() == 0.7853981633974483
            True
            >>> halfedge2.direction() == -2.356194490192345
            True

        Returns:
          The angular direction.
        """
        drct: numpy_array = np.array(
            self.edge.other_vertex(self.vertex).coordinates
        ) - np.array(self.vertex.coordinates)
        norm = np.linalg.norm(drct)
        drct /= norm
        return float(np.arctan2(drct[1], drct[0]))


class Mesh:
    """
    A container that holds a list of unique halfedges.

    Vertices and edges can be retrieved from those.
    The mesh is assumed to be flat (2D).

    """

    def __init__(self, halfedges: list[Halfedge]) -> None:
        """Initialize a Mesh object."""
        self.halfedges = halfedges

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        num = len(self.halfedges)
        return f'Mesh object containing {num} halfedges.'

    def geometric_properties(self) -> dict[str, object]:
        """
        Calculate the geometric properties of the mesh.

        Returns:
          The geometric properties.
        """
        coordinates: numpy_array = np.array(
            [h.vertex.coordinates for h in self.halfedges]
        )
        return geometric_properties(coordinates)

    def bounding_box(self) -> numpy_array:
        """
        Obtain a bounding box of the mesh.

        Returns:
          The bounding box.
        """
        coordinates: numpy_array = np.array(
            [h.vertex.coordinates for h in self.halfedges]
        )
        xmin = min(coordinates[:, 0])
        xmax = max(coordinates[:, 0])
        ymin = min(coordinates[:, 1])
        ymax = max(coordinates[:, 1])
        return np.array([[xmin, ymin], [xmax, ymax]])


############################################
# Geometric Properties of Polygonal Shapes #
############################################


def polygon_area(coordinates: numpy_array) -> float:
    """
    Calculate the area of a polygon.

    Arguments:
        coordinates: A matrix whose columns represent
                the coordinates and the rows
                represent the points of the polygon.
                The first point should not be repeated
                at the end, as this is done
                automatically.

    Returns:
    -------
        area: The area of the polygon.

    Example:
        >>> coordinates = np.array([[0, 0], [1, 0], [1, 1], [0, 1]])
        >>> polygon_area(coordinates)
        1.0

    """
    x_coordinates = coordinates[:, 0]
    y_coordinates = coordinates[:, 1]
    return float(
        np.sum(
            x_coordinates * np.roll(y_coordinates, -1)
            - np.roll(x_coordinates, -1) * y_coordinates
        )
        / 2.00
    )


def polygon_centroid(coordinates: numpy_array) -> numpy_array:
    """
    Calculate the centroid of a polygon.

    Arguments:
        coordinates: A matrix whose columns represent
                the coordinates and the rows
                represent the points of the polygon.
                The first point should not be repeated
                at the end, as this is done
                automatically.

    Returns:
    -------
        centroid: The centroid of the polygon.

    Example:
        >>> coordinates = np.array([[0, 0], [1, 0], [1, 1], [0, 1]])
        >>> polygon_centroid(coordinates)
        array([0.5, 0.5])

    """
    x_coordinates = coordinates[:, 0]
    y_coordinates = coordinates[:, 1]
    area = polygon_area(coordinates)
    x_cent = (
        np.sum(
            (x_coordinates + np.roll(x_coordinates, -1))
            * (
                x_coordinates * np.roll(y_coordinates, -1)
                - np.roll(x_coordinates, -1) * y_coordinates
            )
        )
    ) / (6.0 * area)
    y_cent = (
        np.sum(
            (y_coordinates + np.roll(y_coordinates, -1))
            * (
                x_coordinates * np.roll(y_coordinates, -1)
                - np.roll(x_coordinates, -1) * y_coordinates
            )
        )
    ) / (6.0 * area)
    return np.array((x_cent, y_cent))


def polygon_inertia(coordinates: numpy_array) -> dict[str, float]:
    """
    Calculate the moments of inertia of a polygon.

    Arguments:
        coordinates: A matrix whose columns represent
                the coordinates and the rows
                represent the points of the polygon.
                The first point should not be repeated
                at the end, as this is done
                automatically.

    Returns:
    -------
        dictionary, containing:
        'ixx': (float) - Moment of inertia around
                         the x axis
        'iyy': (float) - Moment of inertia around
                         the y axis
        'ixy': (float) - Product of inertia
        'ir': (float)  - Polar moment of inertia
        'ir_mass': (float) - Mass moment of inertia

    Example:
        >>> coordinates = np.array([[-2, -1], [-2, 1], [1, 1], [1, -1]])
        >>> res = polygon_inertia(coordinates)
        >>> res['ixx'] == -2.0
        True
        >>> res['iyy'] == -6.0
        True
        >>> res['ixy'] == 0.00
        True
        >>> res['ir'] == -8.0
        True
        >>> res['ir_mass'] == 1.33333333333333333333
        True

    """
    x_coordinates = coordinates[:, 0]
    y_coordinates = coordinates[:, 1]
    area = polygon_area(coordinates)
    alpha = (
        x_coordinates * np.roll(y_coordinates, -1)
        - np.roll(x_coordinates, -1) * y_coordinates
    )
    # planar moment of inertia wrt horizontal axis
    ixx = (
        np.sum(
            (
                y_coordinates**2
                + y_coordinates * np.roll(y_coordinates, -1)
                + np.roll(y_coordinates, -1) ** 2
            )
            * alpha
        )
        / 12.00
    )
    # planar moment of inertia wrt vertical axis
    iyy = (
        np.sum(
            (
                x_coordinates**2
                + x_coordinates * np.roll(x_coordinates, -1)
                + np.roll(x_coordinates, -1) ** 2
            )
            * alpha
        )
        / 12.00
    )

    ixy = (
        np.sum(
            (
                x_coordinates * np.roll(y_coordinates, -1)
                + 2.0 * x_coordinates * y_coordinates
                + 2.0 * np.roll(x_coordinates, -1) * np.roll(y_coordinates, -1)
                + np.roll(x_coordinates, -1) * y_coordinates
            )
            * alpha
        )
        / 24.0
    )
    # polar (torsional) moment of inertia
    i_r = ixx + iyy
    # mass moment of inertia wrt in-plane rotation
    ir_mass = (ixx + iyy) / area

    return {'ixx': ixx, 'iyy': iyy, 'ixy': ixy, 'ir': i_r, 'ir_mass': ir_mass}


def geometric_properties(coordinates: numpy_array) -> dict[str, object]:
    """
    Aggregate the results of the previous functions.

    Returns:
      The aggregated results.
    """
    # repeat the first row at the end to close the shape
    coordinates = np.vstack((coordinates, coordinates[0, :]))
    area = polygon_area(coordinates)
    centroid = polygon_centroid(coordinates)
    coordinates_centered = coordinates - centroid
    inertia = polygon_inertia(coordinates_centered)

    return {'area': area, 'centroid': centroid, 'inertia': inertia}


##################################
# Defining halfedges given edges #
##################################

# auxiliary functions


def ang_reduce(ang: float) -> float:
    """
    Brings and angle expressed in radians in the interval [0, 2pi).

    Returns:
      The reduced angle.
    """
    while ang < 0:
        ang += 2.0 * np.pi
    while ang >= 2.0 * np.pi:
        ang -= 2.0 * np.pi
    return ang


def define_halfedges(edges: list[Edge]) -> list[Halfedge]:
    """
    Define all halfedges.

    Given a list of edges, defines all the halfedges and
    associates them with their `next`.

    Note:
        See https://notability.com/n/0wlJ17mt81uuVWAYVoFfV3

        Each halfedge stores information about its edge, vertex and
        and next halfedge. Contrary to convention, we don't store the
        twin (opposite) halfedge here, seince we don't need it
        anywhere.

        This function receives a list of Edge objects as input and
        returns a list of Halfedge objects. The function first creates
        a Halfedge object for each vertex of each Edge object, using
        the `define_halfedge` method of the Edge class. These Halfedge
        objects are stored in a list called all_halfedges. For each
        Halfedge object, the function also updates the list of
        halfedges leaving the vertex that the halfedge originates
        from. For example, if we have two Halfedge objects h1 and h2,
        both originating from the same vertex v, v.halfedges will be a
        list containing h1 and h2. (This is useful because it allows
        us to easily access all the halfedges that originate from a
        particular vertex, which we need later on in the algorithm.)

        After all halfedges have been created, the function assigns
        the next attribute of each halfedge, which points to the next
        halfedge in the sequence. To do this, it loops through all
        halfedges and, for each halfedge h, it determines the vertex
        v_to that h points to, gets a list of all halfedges leaving
        v_to, and assigns the next attribute of h to the halfedge in
        that list that has the smallest angular difference with
        respect to the direction of h.

    Arguments:
        edges: List of Edge objects

    Returns:
    -------
        halfedges: List of Halfedge objects

    Example:
        >>> from osmg.mesh import Vertex, Edge, Halfedge
        >>> # define some vertices
        >>> v1 = Vertex((0.0, 0.0))
        >>> v2 = Vertex((1.0, 0.0))
        >>> v3 = Vertex((1.0, 1.0))
        >>> v4 = Vertex((0.0, 1.0))
        >>> # define some edges
        >>> e1 = Edge(v1, v2)
        >>> e2 = Edge(v2, v3)
        >>> e3 = Edge(v3, v4)
        >>> e4 = Edge(v4, v1)
        >>> # define the halfedges
        >>> halfedges = define_halfedges([e1, e2, e3, e4])
        >>> # check that the `next` attribute of each halfedge
        >>> # is correctly assigned
        >>> for h in halfedges:
        ...     assert h.nxt.vertex == h.edge.other_vertex(h.vertex)

    """
    all_halfedges: list[Halfedge] = []
    for edge in edges:
        v_i = edge.v_i
        v_j = edge.v_j
        h_i = edge.define_halfedge(v_i)
        h_j = edge.define_halfedge(v_j)
        all_halfedges.extend((h_i, h_j))
        v_i.halfedges.append(h_i)
        v_j.halfedges.append(h_j)

    # at this point we have defined all halfedges, but
    # none of them knows its `next`.
    # We now assign that attribute to all halfedges

    for halfedge in all_halfedges:
        # We are looking for `h`'s `next`
        # determine the vertex that it starts from
        v_from = halfedge.vertex
        # determine the vertex that it points to
        v_to = halfedge.edge.other_vertex(v_from)
        # get a list of all halfedges leaving that vertex
        candidates_for_next = v_to.halfedges
        # determine which of all these halfedges will be the next
        angles = np.full(len(candidates_for_next), 0.00)
        for i, h_other in enumerate(candidates_for_next):
            if h_other.edge == halfedge.edge:
                angles[i] = 1000.0
                # otherwise we would assign its conjugate as next
            else:
                angles[i] = ang_reduce(
                    (halfedge.direction() - np.pi) - h_other.direction()
                )
        halfedge.nxt = candidates_for_next[np.argmin(angles)]

    return all_halfedges

    # # debug
    # import matplotlib.pyplot as plt
    # fig = plt.figure()
    # ax = fig.add_subplot(111)
    # ax.set_aspect('equal')
    # for edge in edges:
    #     p1 = edge.v_i.coordinates
    #     p2 = edge.v_j.coordinates
    #     coordinates = np.row_stack((p1, p2))
    #     ax.plot(coordinates[:, 0], coordinates[:, 1])
    # for h in halfedges:
    #     if h.nxt:
    #         h_nxt = h.nxt
    #         e = h.edge
    #         if h_nxt.edge:
    #             e_nxt = h_nxt.edge
    #             p1 = (np.array(e.v_i.coordinates)
    #                   + np.array(e.v_j.coordinates))/2.
    #             p2 = (np.array(e_nxt.v_i.coordinates)
    #                   + np.array(e_nxt.v_j.coordinates))/2.
    #             dx = p2 - p1
    #             ax.arrow(*p1, *dx)
    # plt.show()


def obtain_closed_loops(halfedges: list[Halfedge]) -> list[list[Halfedge]]:
    """
    Obtain closed halfedge loops.

    Given a list of halfedges, this function uses their `next`
    attribute to group them into sequences of closed loops (ordered
    lists of halfedges of which the `next` halfedge of the last list
    element points to the first halfedge in the list, and the `next`
    halfedge of any list element points to the next halfedge in the
    list.

    Arguments:
        halfedges: list of halfedges

    Returns:
    -------
        loops with the aforementioned property.

    """

    def is_in_some_loop(halfedge: Halfedge, loops: list[list[Halfedge]]) -> bool:
        return any(halfedge in loop for loop in loops)

    loops: list[list[Halfedge]] = []
    for halfedge in halfedges:
        if loops:
            if is_in_some_loop(halfedge, loops):
                continue
        loop = [halfedge]
        nxt = halfedge.nxt
        assert nxt
        while nxt != halfedge:
            loop.append(nxt)
            nxt = nxt.nxt
            assert nxt
        loops.append(loop)
    return loops


def orient_loops(
    loops: list[list[Halfedge]],
) -> tuple[
    list[list[Halfedge]],
    list[list[Halfedge]],
    list[list[Halfedge]],
]:
    """
    Orient loops.

    Separates loops to internal (counterclockwise) and external
    (clockwise). Also gathers trivial loops, i.e. halfedge sequences
    that define polygons that have no area (e.g. h1 -> h2 -> h1).

    Arguments:
        loops: (see `obtain_closed_loops`)

    Returns:
    -------
        external_loops (list[list[Halfedge]])
        internal_loops (list[list[Halfedge]])
        trivial_loops (list[list[Halfedge]])

    """
    internal_loops = []
    external_loops = []
    trivial_loops = []
    loop_areas = [
        polygon_area(np.array([h.vertex.coordinates for h in loop]))
        for loop in loops
    ]
    for i, area in enumerate(loop_areas):
        if area > common.EPSILON:
            internal_loops.append(loops[i])
        elif area < -common.EPSILON:
            external_loops.append(loops[i])
        else:
            trivial_loops.append(loops[i])
    return external_loops, internal_loops, trivial_loops


#######################################
# Breaking a shape into little pieces #
#######################################


def subdivide_polygon(
    *, outside: Mesh, holes: list[Mesh], n_x: int, n_y: int, plot: bool = False
) -> list[shapely_Polygon]:
    """
    Define the fibers of fiber sections.

    Arguments:
      outside: Sequence of halfedges that defines the outside shape
      of a section.
      holes: List of sequences of halfedges that define holes.
      n_x: Number of spatial partitions in the x direction
      n_y: Number of spatial partitions in the y direction
      plot: Plots the resulting polygons for debugging

    Returns:
    -------
        pieces: shapely_Polygon objects that represent single fibers.

    """
    outside_polygon = shapely_Polygon(
        [h.vertex.coordinates for h in outside.halfedges]
    )
    hole_polygons = []
    for hole in holes:
        hole_polygons.append(  # noqa: PERF401
            shapely_Polygon([h.vertex.coordinates for h in hole.halfedges])
        )
    remaining_polygon = outside_polygon
    for hole_polygon in hole_polygons:
        remaining_polygon = remaining_polygon.difference(hole_polygon)
    x_min, y_min, x_max, y_max = outside_polygon.bounds
    x_array = np.linspace(x_min, x_max, num=n_x, endpoint=True)
    y_array = np.linspace(y_min, y_max, num=n_y, endpoint=True)
    pieces = []
    for i in range(len(x_array) - 1):
        for j in range(len(y_array) - 1):
            tile = shapely_Polygon(
                [
                    (x_array[i], y_array[j]),
                    (x_array[i + 1], y_array[j]),
                    (x_array[i + 1], y_array[j + 1]),
                    (x_array[i], y_array[j + 1]),
                ]
            )
            subregion = remaining_polygon.intersection(tile)
            if subregion.area != 0.0:
                pieces.append(subregion)
    if plot:
        fig = plt.figure()
        ax_1 = fig.add_subplot(111)
        ax_1.set_aspect('equal')
        patch = PolygonPatch(remaining_polygon, alpha=0.5, zorder=2)
        ax_1.add_patch(patch)
        for subregion in pieces:
            patch = PolygonPatch(subregion, alpha=0.5, zorder=2)
            ax_1.add_patch(patch)
        for subregion in pieces:
            ax_1.scatter(subregion.centroid.x, subregion.centroid.y)
        ax_1.margins(0.10)
        plt.show()
    return pieces


def subdivide_hss_rect(
    sec_h: float, sec_b: float, sec_t: float, *, plot: bool = False
) -> list[shapely_Polygon]:
    """
    Define the fibers of steel HSS fiber sections.

    Arguments:
      sec_h: Section height
      sec_b: Section width
      sec_t: Section thickness
      plot: Whether to create a verification plot, used for debugging
      purposes.

    Returns:
    -------
        pieces: shapely_Polygon objects that represent single fibers.

    """
    outside_polygon = shapely_Polygon(
        np.array(
            (
                (sec_h, sec_b),
                (sec_h, -sec_b),
                (-sec_h, -sec_b),
                (-sec_h, sec_b),
            )
        )
    )
    hole_polygon = shapely_Polygon(
        np.array(
            (
                (sec_h - sec_t, sec_b - sec_t),
                (sec_h - sec_t, -sec_b + sec_t),
                (-sec_h + sec_t, -sec_b + sec_t),
                (-sec_h + sec_t, sec_b - sec_t),
            )
        )
    )
    remaining_polygon = outside_polygon.difference(hole_polygon)
    x_min, y_min, x_max, y_max = outside_polygon.bounds
    # cutting it into 8 regions
    pieces = []
    for ylow, yhigh in zip(
        (y_min, y_min + sec_t, y_max - sec_t),
        (y_min + sec_t, y_max - sec_t, y_max),
    ):
        for xlow, xhigh in zip(
            (x_min, x_min + sec_t, x_max - sec_t),
            (x_min + sec_t, x_max - sec_t, x_max),
        ):
            x_array = np.linspace(xlow, xhigh, num=5, endpoint=True)
            y_array = np.linspace(ylow, yhigh, num=5, endpoint=True)
            for i in range(len(x_array) - 1):
                for j in range(len(y_array) - 1):
                    tile = shapely_Polygon(
                        [
                            (x_array[i], y_array[j]),
                            (x_array[i + 1], y_array[j]),
                            (x_array[i + 1], y_array[j + 1]),
                            (x_array[i], y_array[j + 1]),
                        ]
                    )
                    subregion = remaining_polygon.intersection(tile)
                    if subregion.area != 0.0:
                        pieces.append(subregion)

    if plot:
        fig = plt.figure()
        ax_1 = fig.add_subplot(111)
        ax_1.set_aspect('equal')
        # patch = PolygonPatch(remaining_polygon, alpha=0.5, zorder=2)
        # ax_1.add_patch(patch)
        for subregion in pieces:
            patch = PolygonPatch(subregion, alpha=0.5, zorder=2)
            ax_1.add_patch(patch)
        for subregion in pieces:
            ax_1.scatter(subregion.centroid.x, subregion.centroid.y)
        ax_1.margins(0.10)
        plt.show()

    return pieces


def subdivide_hss_circ(
    sec_d: float, sec_t: float, *, plot: bool = False
) -> list[shapely_Polygon]:
    """
    Define the fibers of steel HSS fiber sections.

    Arguments:
      sec_d: Section diameter
      sec_t: Section thickness
      plot: Whether to create a verification plot, used for debugging
      purposes.

    Returns:
    -------
        pieces: shapely_Polygon objects that represent single fibers.

    """
    num_subdiv_t = 3
    num_subdiv_circ = 12

    radius = sec_d / 2.00

    pieces = []

    for i in range(num_subdiv_t):
        for j in range(num_subdiv_circ):
            rr_i = radius - i * sec_t / num_subdiv_t
            rr_j = radius - (i + 1.0) * sec_t / num_subdiv_t
            ang_i = (2.00 * np.pi) / num_subdiv_circ * j
            ang_j = (2.00 * np.pi) / num_subdiv_circ * (j + 1.00)
            pt1 = (rr_i * np.cos(ang_i), rr_i * np.sin(ang_i))
            pt2 = (rr_i * np.cos(ang_j), rr_i * np.sin(ang_j))
            pt3 = (rr_j * np.cos(ang_j), rr_j * np.sin(ang_j))
            pt4 = (rr_j * np.cos(ang_i), rr_j * np.sin(ang_i))
            pol = shapely_Polygon((pt1, pt2, pt3, pt4))
            pieces.append(pol)

    if plot:
        _, ax = plt.subplots()
        ax.set_aspect('equal')
        for piece in pieces:
            patch = Polygon(piece.exterior.coordinates, alpha=0.5, zorder=2)
            ax.add_patch(patch)
        for piece in pieces:
            ax.scatter(piece.centroid.x, piece.centroid.y)
        ax.margins(0.10)
        plt.show()

    return pieces


#############
# Debugging #
#############


def print_halfedge_results(halfedges: list[Halfedge]) -> None:
    """
    Print the ids of defined halfedges.

    Prints the ids of the defined halfedges and their vertex, edge and
    next, for debugging.

    """
    results: dict[str, list[Any]] = {
        'halfedge': [],
        'vertex': [],
        'edge': [],
        'next': [],
    }

    for halfedge in halfedges:
        results['halfedge'].append(halfedge)
        results['vertex'].append(halfedge.vertex)
        results['edge'].append(halfedge.edge)
        results['next'].append(halfedge.nxt)

    print(results)  # noqa: T201


def plot_loop(halfedge_loop: list[Halfedge]) -> None:
    """Plot the vertices/edges of a list of halfedges."""
    num = len(halfedge_loop)
    coordinates = np.full((num + 1, 2), 0.00)
    for i, halfedge in enumerate(halfedge_loop):
        coordinates[i, :] = halfedge.vertex.coordinates
    coordinates[-1, :] = coordinates[0, :]
    fig = plt.figure()
    plt.plot(coordinates[:, 0], coordinates[:, 1])
    plt.scatter(coordinates[:, 0], coordinates[:, 1])
    fig.show()


def plot_edges(edges: list[Edge]) -> None:
    """Plot the given edges."""
    fig = plt.figure()
    for edge in edges:
        coordinates = np.full((2, 2), 0.00)
        coordinates[0, :] = edge.v_i.coordinates
        coordinates[1, :] = edge.v_j.coordinates
        plt.plot(coordinates[:, 0], coordinates[:, 1])
    fig.show()


def sanity_checks(
    external: list[list[Halfedge]], trivial: list[list[Halfedge]]
) -> None:
    """Perform some checks to make sure assumptions are not violated."""
    #   We expect no trivial loops
    if trivial:
        print('Warning: Found trivial loop')  # noqa: T201
        for trv in trivial:
            for halfedge in trv:
                print(halfedge.vertex.coordinates)  # noqa: T201
            plot_loop(trv)
    #   We expect a single external loop
    if len(external) > 1:
        print('Warning: Found multiple external loops')  # noqa: T201
        for i, ext in enumerate(external):
            print(i + 1)  # noqa: T201
            for halfedge in ext:
                print(halfedge.vertex.coordinates)  # noqa: T201
            plot_loop(ext)


if __name__ == '__main()__':
    pass



"""Generates meshes for preconfigured sections."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from osmg.geometry.mesh import (
    Edge,
    Mesh,
    Vertex,
    define_halfedges,
    obtain_closed_loops,
    orient_loops,
    sanity_checks,
)

numpy_array = npt.NDArray[np.float64]


def generate(edges: list[Edge]) -> Mesh:
    """
    Generate a mesh from the given edges.

    Returns:
      The generated mesh.
    """
    halfedges = define_halfedges(edges)
    loops = obtain_closed_loops(halfedges)
    _, internal, trivial = orient_loops(loops)
    sanity_checks(internal, trivial)
    return Mesh(internal[0])


def define_edges(vertices: list[Vertex]) -> list[Edge]:
    """
    Define edges from an ordered list of vertices.

    Returns:
      The defined edges.
    """
    n_v = len(vertices)
    edges = []
    for i in range(n_v - 1):
        v_i = vertices[i]
        v_j = vertices[i + 1]
        edges.append(Edge(v_i, v_j))
    v_i = vertices[-1]
    v_j = vertices[0]
    edges.append(Edge(v_i, v_j))
    return edges


def w_mesh(
    sec_b: float,
    sec_h: float,
    sec_tw: float,
    sec_tf: float,
    target_area: float,
) -> Mesh:
    """
    W-section mesh.

    Defines a loop of counterclockwise halfedges that form the shape
    of the W section with the specified parameters.  The origin
    coincides with the centroid.

    Arguments:
      sec_b: total width
      sec_h: total height
      sec_tw: web thickness
      sec_tf: flange thickness
      target_area: AISC database area to determine fillets, because
        trying to do that using `T` doesn't work.

    Returns:
      The generated mesh.
    """
    area_diff = target_area - (sec_b * sec_tf * 2.0 + (sec_h - 2 * sec_tf) * sec_tw)
    if area_diff < 0:
        # This happens for W14X426
        area_diff = 1e-4

    dist = np.sqrt(area_diff / (2.0**2 - np.pi)) * 0.9565
    # note: 0.9565 is a correction factor to account for
    # the fact that we approximate the arcs with
    # four line segments, thus putting more material in there
    k = (sec_b - 2.0 * dist - sec_tw) / 2.0
    vertices = [
        Vertex((sec_b / 2.0, sec_h / 2.0)),
        Vertex((-sec_b / 2.0, sec_h / 2.0)),
        Vertex((-sec_b / 2.0, sec_h / 2.0 - sec_tf)),
        Vertex((-sec_b / 2.0 + k, sec_h / 2.0 - sec_tf)),
        Vertex(
            (
                -sec_b / 2.0 + k + dist * np.cos(3.0 * np.pi / 8.0),
                sec_h / 2.0 - sec_tf - dist + dist * np.sin(3.0 * np.pi / 8.0),
            )
        ),
        Vertex(
            (
                -sec_b / 2.0 + k + dist * np.cos(1.0 * np.pi / 4.0),
                sec_h / 2.0 - sec_tf - dist + dist * np.sin(1.0 * np.pi / 4.0),
            )
        ),
        Vertex(
            (
                -sec_b / 2.0 + k + dist * np.cos(1.0 * np.pi / 8.0),
                sec_h / 2.0 - sec_tf - dist + dist * np.sin(1.0 * np.pi / 8.0),
            )
        ),
        Vertex((-sec_b / 2.0 + k + dist, sec_h / 2.0 - sec_tf - dist)),
        Vertex((-sec_b / 2.0 + k + dist, -sec_h / 2.0 + sec_tf + dist)),
        Vertex(
            (
                -sec_b / 2.0 + k + dist * np.cos(1.0 * np.pi / 8.0),
                -sec_h / 2.0 + sec_tf + dist - dist * np.sin(1.0 * np.pi / 8.0),
            )
        ),
        Vertex(
            (
                -sec_b / 2.0 + k + dist * np.cos(1.0 * np.pi / 4.0),
                -sec_h / 2.0 + sec_tf + dist - dist * np.sin(1.0 * np.pi / 4.0),
            )
        ),
        Vertex(
            (
                -sec_b / 2.0 + k + dist * np.cos(3.0 * np.pi / 8.0),
                -sec_h / 2.0 + sec_tf + dist - dist * np.sin(3.0 * np.pi / 8.0),
            )
        ),
        Vertex((-sec_b / 2.0 + k, -sec_h / 2.0 + sec_tf)),
        Vertex((-sec_b / 2.0, -(sec_h / 2.0 - sec_tf))),
        Vertex((-sec_b / 2.0, -sec_h / 2.0)),
        Vertex((sec_b / 2.0, -sec_h / 2.0)),
        Vertex((sec_b / 2.0, -(sec_h / 2 - sec_tf))),
        Vertex((+sec_b / 2.0 - k, -sec_h / 2.0 + sec_tf)),
        Vertex(
            (
                +sec_b / 2.0 - k - dist * np.cos(3.0 * np.pi / 8.0),
                -sec_h / 2.0 + sec_tf + dist - dist * np.sin(3.0 * np.pi / 8.0),
            )
        ),
        Vertex(
            (
                +sec_b / 2.0 - k - dist * np.cos(1.0 * np.pi / 4.0),
                -sec_h / 2.0 + sec_tf + dist - dist * np.sin(1.0 * np.pi / 4.0),
            )
        ),
        Vertex(
            (
                +sec_b / 2.0 - k - dist * np.cos(1.0 * np.pi / 8.0),
                -sec_h / 2.0 + sec_tf + dist - dist * np.sin(1.0 * np.pi / 8.0),
            )
        ),
        Vertex((+sec_b / 2.0 - k - dist, -sec_h / 2.0 + sec_tf + dist)),
        Vertex((+sec_b / 2.0 - k - dist, +sec_h / 2.0 - sec_tf - dist)),
        Vertex(
            (
                +sec_b / 2.0 - k - dist * np.cos(1.0 * np.pi / 8.0),
                +sec_h / 2.0 - sec_tf - dist + dist * np.sin(1.0 * np.pi / 8.0),
            )
        ),
        Vertex(
            (
                +sec_b / 2.0 - k - dist * np.cos(1.0 * np.pi / 4.0),
                +sec_h / 2.0 - sec_tf - dist + dist * np.sin(1.0 * np.pi / 4.0),
            )
        ),
        Vertex(
            (
                +sec_b / 2.0 - k - dist * np.cos(3.0 * np.pi / 8.0),
                +sec_h / 2.0 - sec_tf - dist + dist * np.sin(3.0 * np.pi / 8.0),
            )
        ),
        Vertex((+sec_b / 2.0 - k, sec_h / 2.0 - sec_tf)),
        Vertex((sec_b / 2.0, sec_h / 2.0 - sec_tf)),
    ]
    edges = define_edges(vertices)
    return generate(edges)


def rect_mesh(dim_b: float, dim_h: float) -> Mesh:
    """
    Rectangular mesh.

    Defines a loop of counterclockwise halfedges
    that form the shape of the rectangular section with
    the specified parameters.
    The origin coincides with the centroid.

    Arguments:
        dim_b: total width
        dim_h: total height

    Returns:
      The generated mesh.
    """
    vertices = [
        Vertex((dim_b / 2.0, dim_h / 2.0)),
        Vertex((-dim_b / 2.0, dim_h / 2.0)),
        Vertex((-dim_b / 2.0, -dim_h / 2.0)),
        Vertex((dim_b / 2.0, -dim_h / 2.0)),
    ]
    edges = define_edges(vertices)
    return generate(edges)


def circ_mesh(dim_d: float) -> Mesh:
    """
    Circular mesh.

    Defines a loop of counterclockwise halfedges
    that form the shape of the circular section with
    the specified parameters.
    The origin coincides with the centroid.

    Arguments:
        dim_d: total diameter

    Returns:
      The generated mesh.
    """
    radius = dim_d / 2.0
    num_vertices = 32  # Number of vertices on the circumference

    angle_increment = 2 * np.pi / num_vertices

    vertices = []
    for i in range(num_vertices):
        angle = i * angle_increment
        vertices.append(Vertex((radius * np.cos(angle), radius * np.sin(angle))))

    edges = define_edges(vertices)
    return generate(edges)


def generic_snap_points(mesh: Mesh) -> dict[str, numpy_array]:
    """
    Generate generic snap points for a section object.

    Returns:
      The snap points.
    """
    bbox = mesh.bounding_box()
    z_min, y_min, z_max, y_max = bbox.flatten()
    snap_points: dict[str, numpy_array] = {}
    snap_points['centroid'] = -np.array([0.0, 0.0])
    snap_points['top_center'] = -np.array([0.0, y_max])
    snap_points['top_left'] = -np.array([z_min, y_max])
    snap_points['top_right'] = -np.array([z_max, y_max])
    snap_points['center_left'] = -np.array([z_min, 0.0])
    snap_points['center_right'] = -np.array([z_max, 0.0])
    snap_points['bottom_center'] = -np.array([0.0, y_min])
    snap_points['bottom_left'] = -np.array([z_min, y_min])
    snap_points['bottom_right'] = -np.array([z_max, y_min])
    return snap_points



"""Defines :obj:`~osmg.line.Line` objects."""

from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt

from osmg.core import common

numpy_array = npt.NDArray[np.float64]


@dataclass
class Line:
    """
    Finite-length line segment object.

    Used internally whenever operations involving lines are required.

    Attributes:
    ----------
      tag: line tag.
      start: starting point.
      end: end point.

    """

    tag: str
    start: numpy_array = field(repr=False)
    end: numpy_array = field(repr=False)

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'Line object\n'
        res += f'  start: {self.start}\n'
        res += f'  end: {self.end}\n'
        return res

    def length(self) -> float:
        """
        Obtain the length of the line.

        Example:
            >>> from osmg.line import Line
            >>> l1 = Line('l1', np.array([0, 0]), np.array([2, 2]))
            >>> l1.length() == 2.8284271247461903
            True

        Returns:
          The string representation of the object.
        """
        return float(np.linalg.norm(self.end - self.start))

    def direction(self) -> numpy_array:
        """
        Line direction.

        Returns a unit vector pointing from the start to the end of
        the line.

        Example:
            >>> from osmg.line import Line
            >>> l1 = Line('l1', np.array([0, 0]), np.array([2, 2]))
            >>> l1.direction()
            array([0.70710678, 0.70710678])

        Returns:
          The string representation of the object.
        """
        return (self.end - self.start) / self.length()

    def intersect(self, other: Line) -> numpy_array | None:
        """
        Intersection point.

        Calculates the intersection point of this line with another
        line. Returns None if the lines don't intersect.  Note: 'line'
        is actually a finite-length line segment.

        Parameters
        ----------
          other: the other line

        Example:
            >>> from osmg.line import Line
            >>> l1 = Line('l1', np.array([0, 0]), np.array([2, 2]))
            >>> l2 = Line('l2', np.array([1, 0]), np.array([1, 3]))
            >>> l1.intersect(l2)
            array([1., 1.])

        Returns:
          The intersection point if it exists.
        """
        ra_dir = self.direction()
        rb_dir = other.direction()
        mat: numpy_array = np.array(
            [[ra_dir[0], -rb_dir[0]], [ra_dir[1], -rb_dir[1]]]
        )
        if np.abs(np.linalg.det(mat)) <= common.EPSILON:
            # The lines are parallel
            # in this case, we check if they have
            # a common starting or ending point
            # (we ignore the case of a common segment,
            #  as it has no practical use for our purposes).
            if (
                np.linalg.norm(self.start - other.start) <= common.EPSILON
                or np.linalg.norm(self.start - other.end) <= common.EPSILON
            ):
                return self.start
            if (
                np.linalg.norm(self.end - other.start) <= common.EPSILON
                or np.linalg.norm(self.end - other.end) <= common.EPSILON
            ):
                return self.end
            return None
        # Get the origins
        ra_ori = self.start
        rb_ori = other.start
        # System left-hand-side
        bvec: numpy_array = np.array(
            [
                [rb_ori[0] - ra_ori[0]],
                [rb_ori[1] - ra_ori[1]],
            ]
        )
        # Solve to get u and v in a vector
        uvvec = np.linalg.solve(mat, bvec)
        # Terminate if the intersection point
        # does not lie on both lines
        if uvvec[0] < 0 - common.EPSILON:
            return None
        if uvvec[1] < 0 - common.EPSILON:
            return None
        if uvvec[0] > self.length() + common.EPSILON:
            return None
        if uvvec[1] > other.length() + common.EPSILON:
            return None
        # Otherwise the point is valid
        point = ra_ori + ra_dir * uvvec[0]
        return np.array([point[0], point[1]])

    def intersects_pt(self, point: numpy_array) -> bool:
        """
        Check whether the given point pt lies on the line.

        Parameters
        ----------
            point: a point

        Returns: True if the point lies on the line, False otherwise

        Example:
            >>> from osmg.line import Line
            >>> l = Line('my_line', np.array([0, 0]), np.array([1, 1]))
            >>> l.intersects_pt(np.array([0.5, 0.5]))
            True
            >>> l.intersects_pt(np.array([0, 0]))
            True
            >>> l.intersects_pt(np.array([1, 1]))
            True
            >>> l.intersects_pt(np.array([2, 2]))
            False

        Returns:
          Whether the given point pt lies on the line.

        Raises:
          ValueError: If the line has zero length.
        """
        r_a = self.end - self.start
        norm2 = np.dot(r_a, r_a)
        if np.abs(norm2) < common.EPSILON:
            msg = 'Line has zero length.'
            raise ValueError(msg)
        r_b = point - self.start

        r_a_3d = np.append(r_a, 0)
        r_b_3d = np.append(r_b, 0)
        cross = np.linalg.norm(np.cross(r_a_3d, r_b_3d))

        dot_normalized = np.dot(r_a, r_b) / norm2  # type: ignore
        if cross < common.EPSILON:
            res = bool(0.00 <= dot_normalized <= 1.00)
        else:
            res = False
        return res

    def point_distance(self, point: numpy_array) -> float | None:
        """
        Minimum distance.

        Calculate the minimum distance between the line segment and a
        point.  If the point falls on the line but is outside of the
        line segment, returns None.

        Parameters:
            point: the point

        Returns:
          The minimum distance.

        Example:
            >>> line = Line(tag='line',
            ...             start=np.array([1, 1]),
            ...             end=np.array([3, 3]))
            >>> point = np.array([4, 2])
            >>> line.point_distance(point)
            1.4142135623730951
            >>> point = np.array([2, 2])
            >>> line.point_distance(point)
            0.0
            >>> point = np.array([0, 0])
            >>> line.point_distance(point)

            >>> point = np.array([4, 4])
            >>> line.point_distance(point)

        """
        r_a = self.end - self.start
        r_b = point - self.start
        proj_point = (r_b @ r_a) / (r_a @ r_a) * r_a
        if self.intersects_pt(proj_point + self.start):
            res: float | None = float(np.linalg.norm(r_b - proj_point))
        else:
            res = None
        return res

    def project(self, point: numpy_array) -> numpy_array | None:
        """
        Projection.

        Calculates the projection of a point on the line.
        If the projection falls on the line segment, it returns the
        projected point, otherwise it returns None.

        Arguments:
          point: the point's coordinates

        Example:
            >>> line = Line('test', np.array([0, 0]), np.array([10, 0]))
            >>> line.project(np.array([5, 0]))
            array([5., 0.])
            >>> line.project(np.array([5, 5]))
            array([5., 0.])
            >>> line.project(np.array([-5, 5]))

            >>> line.project(np.array([15, 5]))

        Returns:
          The projection point if it exists.
        """
        r_a = self.end - self.start
        r_b = point - self.start
        proj_point: numpy_array = (r_b @ r_a) / (r_a @ r_a) * r_a + self.start
        if self.intersects_pt(proj_point):
            return proj_point
        return None



"""Defines :obj:`~osmg.model_objects.node.Node` objects."""

from __future__ import annotations

from dataclasses import dataclass
from functools import total_ordering
from typing import Self

from osmg.core.uid_object import UIDObject


@total_ordering
@dataclass()
class Node(UIDObject):
    """
    OpenSees node.

    https://openseespydoc.readthedocs.io/en/latest/src/node.html?highlight=node

    Attributes:
    ----------
        uid_generator: Unique ID generator object.
        coordinates: List of node coordinates.
        uid: Unique ID of the node, assigned using the generator object.
        restraint: List of boolean values identifying whether the
          corresponding DOF is restrained.
    """

    coordinates: tuple[float, ...]

    def __le__(self, other: Self) -> bool:
        """
        Less or equal determination rule.

        Returns:
          The outcome of the less or equal operation.
        """
        return self.uid <= other.uid

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'Node object\n'
        res += f'  uid: {self.uid}\n'
        res += f'  coordinates: {self.coordinates}\n'
        return res

    def __hash__(self) -> int:
        """Return the hash of the object based on its UID."""
        return hash(self.uid)



"""Unit tests for gridsystem."""

import numpy as np
import pytest

from osmg.core.gridsystem import BaseGridSystem, GridSystem, GridSystem2D
from osmg.geometry.line import Line


class TestBaseGridSystem:
    """Tests for the BaseGridSystem class."""

    def test_add_and_retrieve_level(self) -> None:
        """Test adding and retrieving a level."""
        grid_system = BaseGridSystem[str]()
        grid_system.add_level('Ground Floor', 0.0)
        assert grid_system.get_level('Ground Floor').elevation() == 0.0  # type: ignore

    def test_retrieve_nonexistent_level(self) -> None:
        """Test retrieving a nonexistent level raises ValueError."""
        grid_system = BaseGridSystem[str]()
        with pytest.raises(ValueError, match="Level 'Basement' does not exist."):
            grid_system.get_level('Basement').elevation()  # type: ignore


class TestGridSystem:
    """Tests for the GridSystem class."""

    def test_add_and_retrieve_grid(self) -> None:
        """Test adding and retrieving a grid."""
        grid_system = GridSystem()
        start = np.array([0.0, 0.0])
        end = np.array([10.0, 0.0])
        grid_system.add_grid('Grid A', start, end)
        assert isinstance(grid_system.grids['Grid A'], Line)
        assert np.array_equal(grid_system.grids['Grid A'].start, start)
        assert np.array_equal(grid_system.grids['Grid A'].end, end)

    def test_find_intersection(self) -> None:
        """Test finding the intersection of two grids."""
        grid_system = GridSystem()
        grid_system.add_grid('Grid A', np.array([0.0, 0.0]), np.array([10.0, 0.0]))
        grid_system.add_grid('Grid B', np.array([5.0, -5.0]), np.array([5.0, 5.0]))
        intersection = grid_system.get_intersection_coordinates('Grid A', 'Grid B')
        assert intersection is not None
        assert np.array_equal(intersection, np.array([5.0, 0.0]))

    def test_intersection_with_nonexistent_grid(self) -> None:
        """Test finding the intersection with a nonexistent grid raises ValueError."""
        grid_system = GridSystem()
        grid_system.add_grid('Grid A', np.array([0.0, 0.0]), np.array([10.0, 0.0]))
        with pytest.raises(
            ValueError, match="Grids 'Grid A' and 'Grid C' must exist."
        ):
            grid_system.get_intersection_coordinates('Grid A', 'Grid C')


class TestGridSystem2D:
    """Tests for the GridSystem2D class."""

    def test_add_and_retrieve_grid(self) -> None:
        """Test adding and retrieving a grid."""
        grid_system = GridSystem2D()
        grid_system.add_grid('Grid X', 10.0)
        assert grid_system.get_grid_location('Grid X') == 10.0

    def test_retrieve_nonexistent_grid(self) -> None:
        """Test retrieving a nonexistent grid raises ValueError."""
        grid_system = GridSystem2D()
        with pytest.raises(ValueError, match="Grid 'Grid Y' does not exist."):
            grid_system.get_grid_location('Grid Y')

    def test_add_and_retrieve_level(self) -> None:
        """Test adding and retrieving a level."""
        grid_system = GridSystem2D()
        grid_system.add_level('Ground Floor', 0.0)
        assert grid_system.get_level('Ground Floor').elevation() == 0.0  # type: ignore

    def test_retrieve_nonexistent_level(self) -> None:
        """Test retrieving a nonexistent level raises ValueError."""
        grid_system = GridSystem2D()
        with pytest.raises(ValueError, match="Level 'Roof' does not exist."):
            grid_system.get_level('Roof').elevation()  # type: ignore



"""Defines Analysis objects."""

from __future__ import annotations

from time import perf_counter
from osmg.core import common
import logging
import contextlib
import platform
import socket
import sys
import tempfile
from dataclasses import dataclass, field
from itertools import product
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
from tqdm import tqdm

from osmg.analysis.load_case import HasLoads, ModalLoadCase, StaticLoadCase
from osmg.analysis.recorders import ElementRecorder, NodeRecorder
from osmg.core.common import NDF, NDM, THREE_DIMENSIONAL, TWO_DIMENSIONAL
from osmg.core.model import Model
from osmg.core.osmg_collections import BarAssembly, BeamColumnAssembly
from osmg.model_objects.element import (
    Bar,
    BeamColumnElement,
    DispBeamColumn,
    ElasticBeamColumn,
    TwoNodeLink,
    ZeroLength,
    LeadRubberX,
)

try:
    import opensees.openseespy as ops
except (ImportError, ModuleNotFoundError):
    import openseespy.opensees as ops

if TYPE_CHECKING:
    from osmg.analysis.common import UDL, PointLoad, PointMass
    from osmg.analysis.load_case import LoadCaseRegistry
    from osmg.analysis.recorders import Recorder
    from osmg.analysis.supports import ElasticSupport, FixedSupport
    from osmg.core.model import Model
    from osmg.core.osmg_collections import ComponentAssembly
    from osmg.model_objects.uniaxial_material import UniaxialMaterial


@dataclass()
class AnalysisSettings:
    """Analysis settings object."""

    result_directory: str | None = field(default=None)
    log_file_name: str = field(default='log.txt')
    log_level: int = field(default=logging.DEBUG)
    solver: str = field(default='UmfPack')
    restrict_dof: tuple[bool, ...] | None = field(default=None)
    disable_default_recorders: bool = field(default=False)
    constraints: tuple[str] = ('Transformation',)
    numberer: Literal['RCM'] = 'RCM'
    system: Literal['Umfpack'] = 'Umfpack'
    ignore_by_tag: set[str] = field(default_factory=set)


@dataclass()
class TransientDriftCheckSetup:
    """Transient analysis drift check setup."""

    node_uids: list[int]
    drift_limit: float
    elevation_dof: int = field(default=3)


@dataclass(repr=False)  # noqa: PLR0904
class Analysis:
    """Parent analysis class."""

    model: Model = field(default_factory=Model)
    settings: AnalysisSettings = field(default_factory=AnalysisSettings)
    _logger: logging.Logger = field(init=False)
    recorders: dict[str, Recorder] = field(default_factory=dict)
    _defined_materials: list[int] = field(default_factory=list)
    _basic_force_cache: dict = field(default_factory=dict, init=False)
    _time_series_tags: list = field(default_factory=list)
    _yielded_elements: list = field(default_factory=list)

    def initialize_logger(self) -> None:
        """
        Initialize the logger of the analysis.

        Analysis objects are instantiated with a default configuration
        to make it possible for each load case to initialize a
        corresponding analysis. Their configuration needs to be
        updated before running. Part of that is to set the log file
        path.

        Raises:
          ValueError: If the log file path is still None.
        """
        if (
            self.settings.log_file_name is None
            or self.settings.result_directory is None
        ):
            msg = 'Analysis log file is required.'
            raise ValueError(msg)
        logging.basicConfig(
            filename=(
                Path(self.settings.result_directory) / self.settings.log_file_name
            ),
            filemode='w',
            format='%(asctime)s [%(levelname)s] %(message)s',
            datefmt='%m/%d/%Y %I:%M:%S %p',
            force=True,
        )
        self._logger: logging.Logger = logging.getLogger('OpenSees_Model_Generator')
        self._logger.setLevel(self.settings.log_level)

        # Log system info
        self._logger.info('Analysis initialized.')
        os_system = platform.system()
        self._logger.info(f'Platform: {os_system}.')
        if os_system == 'Linux':
            self._logger.info(f'Hostname: {socket.gethostname()}.')
        self._logger.info(f'Python Version: {sys.version}.')

    def log(
        self,
        msg: str,
        level: Literal['debug', 'info', 'warning', 'error', 'critical'] = 'info',
    ) -> None:
        """
        Add a message to the log file.

        Raises:
          ValueError: If an invalid logging level is specified.
        """
        if level == 'info':
            self._logger.info(msg)
        elif level == 'debug':
            self._logger.debug(msg)
        elif level == 'warning':
            self._logger.warning(msg)
        elif level == 'error':
            self._logger.error(msg)
        elif level == 'critical':
            self._logger.critical(msg)
        else:
            msg = f'Invalid logging level: {level}'
            raise ValueError(msg)

    def opensees_instantiate(self) -> None:
        """Instantiate the model in OpenSees."""
        ops.wipe()
        ops.model(
            'basic',
            '-ndm',
            NDM[self.model.dimensionality],
            '-ndf',
            NDF[self.model.dimensionality],
        )

    def opensees_define_material(self, material: UniaxialMaterial) -> None:
        """Recursively define materials with predecessors."""
        if material.uid not in self._defined_materials:
            while (
                hasattr(material, 'predecessor')
                and material.predecessor.uid not in self._defined_materials
            ):
                self.opensees_define_material(material.predecessor)
            ops.uniaxialMaterial(*material.ops_args())
            self._defined_materials.append(material.uid)

    def opensees_define_nodes(self) -> None:
        """Define the nodes of the model in OpenSees."""
        for uid, node in self.model.get_all_nodes(
            self.settings.ignore_by_tag
        ).items():
            ops.node(uid, *node.coordinates)

    def opensees_define_elements(self) -> None:
        """Define elements."""
        elastic_beamcolumn_elements: list[ElasticBeamColumn] = []
        bar_elements: list[Bar] = []
        two_node_link_elements: list[TwoNodeLink] = []
        zerolength_elements: list[ZeroLength] = []
        lead_rubber_x_elements: list[ZeroLength] = []
        unsupported_element_types: list[str] = []

        # Note: Materials are defined on an as-needed basis.  We keep
        # track of defined materials in `self._defined_materials` to
        # avoid trying to define the same material twice. After
        # elements are defined, we clear that list, to start with an
        # empty list in subsequent model definitions.

        components = self.model.components.values()
        for component in components:
            if component.tags & self.settings.ignore_by_tag:
                continue
            elements = component.elements
            for element in elements.values():
                if isinstance(element, ElasticBeamColumn):
                    elastic_beamcolumn_elements.append(element)
                elif isinstance(element, Bar):
                    bar_elements.append(element)
                elif isinstance(element, TwoNodeLink):
                    two_node_link_elements.append(element)
                elif isinstance(element, ZeroLength):
                    zerolength_elements.append(element)
                elif isinstance(element, LeadRubberX):
                    lead_rubber_x_elements.append(element)
                else:
                    unsupported_element_types.append(element.__class__.__name__)

        if unsupported_element_types:
            print(  # noqa: T201
                f'WARNING: Unsupported element types found: {set(unsupported_element_types)}'
            )

        self.opensees_define_elastic_beamcolumn_elements(elastic_beamcolumn_elements)
        self.opensees_define_bar_elements(bar_elements)
        self.opensees_define_two_node_link_elements(two_node_link_elements)
        self.opensees_define_zerolength_elements(zerolength_elements)
        self.opensees_define_lead_rubber_x_elements(lead_rubber_x_elements)

        # clear defined materials
        self._defined_materials = []

    def opensees_define_lead_rubber_x_elements(
        self, elements: list[LeadRubberX]
    ) -> None:
        """
        Define LeadRubberX elements.

        Raises:
          ValueError: If the analysis is not 3D.
        """
        ndm = NDM[self.model.dimensionality]
        ndf = NDF[self.model.dimensionality]
        if not (ndm == 3 and ndf == 6):
            msg = 'LeadRubberX elements only work with ndm=3 and ndf=6.'
            raise ValueError(msg)
        for element in elements:
            ops.element(*element.ops_args())

    def opensees_define_elastic_beamcolumn_elements(
        self, elements: list[ElasticBeamColumn]
    ) -> None:
        """
        Define elastic beamcolumn elements.

        Raises:
          TypeError: If the model type is invalid.
        """
        for element in elements:
            if element.visibility.skip_opensees_definition:
                continue
            ops.geomTransf(*element.geomtransf.ops_args())
            if self.model.dimensionality == '2D Frame':
                ops.element(*element.ops_args_2d())
            elif self.model.dimensionality == '3D Frame':
                ops.element(*element.ops_args())
            else:
                msg = f'Invalid model dimensionality: `{self.model.dimensionality}`.'
                raise TypeError(msg)

    def opensees_define_bar_elements(self, elements: list[Bar]) -> None:
        """Define bar elements."""
        for element in elements:
            self.opensees_define_material(element.material)
            ops.element(*element.ops_args())

    def opensees_define_two_node_link_elements(
        self, elements: list[TwoNodeLink]
    ) -> None:
        """Define TwoNodeLink elements."""
        for element in elements:
            for material in element.materials:
                self.opensees_define_material(material)
            ops.element(*element.ops_args())

    def opensees_define_zerolength_elements(
        self, elements: list[ZeroLength]
    ) -> None:
        """Define Zerolength elements."""
        for element in elements:
            for material in element.materials:
                self.opensees_define_material(material)
            ops.element(*element.ops_args())

    def opensees_define_node_restraints(
        self,
        fixed_supports: dict[int, FixedSupport],
        elastic_supports: dict[int, ElasticSupport],
    ) -> None:
        """Define node restraints."""
        ndf = NDF[self.model.dimensionality]

        for uid, support in fixed_supports.items():
            fix = []
            for i in range(ndf):
                if support[i] is True or (
                    self.settings.restrict_dof
                    and self.settings.restrict_dof[i] is True
                ):
                    fix.append(True)
                else:
                    fix.append(False)
            if True in fix:
                ops.fix(uid, *[int(x) for x in fix])

        nodes = self.model.get_all_nodes()
        elastic_materials = {}
        for uid, support in elastic_supports.items():
            assert len(support) == ndf
            node = nodes[uid]
            # for each direction.
            material_uids_for_this_support = []
            for value in support:
                # define material if needed.
                if value not in elastic_materials:
                    material_uid = next(self.model.uid_generator.MATERIAL)
                    elastic_materials[value] = material_uid
                    ops.uniaxialMaterial('Elastic', material_uid, value)
                else:
                    material_uid = elastic_materials[value]
                material_uids_for_this_support.append(material_uid)
            # define a node at the same location.
            new_node_uid = next(self.model.uid_generator.NODE)
            ops.node(new_node_uid, *node.coordinates)
            # fix that node.
            ops.fix(new_node_uid, *([1] * ndf))
            # define a zerolength element connecting the two nodes.
            ops.element(
                'zeroLength',
                next(self.model.uid_generator.ELEMENT),
                uid,
                new_node_uid,
                '-mat',
                *material_uids_for_this_support,
                '-dir',
                *range(1, ndf + 1),
            )

    def opensees_define_node_constraints(
        self, rigid_diaphragm: dict[int, tuple[int, ...]]
    ) -> None:
        """
        Define node constraints.

        Raises:
          ValueError: If the model dimensionality is not supported.
        """
        if not rigid_diaphragm:
            return

        for (
            parent_node_uid,
            children_node_uids,
        ) in rigid_diaphragm.items():
            if self.model.dimensionality in {'3D Frame', '3D Truss'}:
                ops.rigidDiaphragm(3, parent_node_uid, *children_node_uids)
            elif self.model.dimensionality in {'2D Frame', '2D Truss'}:
                for child_node_uid in children_node_uids:
                    ops.equalDOF(parent_node_uid, child_node_uid, 1)
            else:
                msg = 'Unsupported model dimensionality: {model.dimensionality}'
                raise ValueError(msg)

    def opensees_define_model(self) -> None:
        """Define the model in OpenSees."""
        self.opensees_instantiate()
        self.opensees_define_nodes()
        self.opensees_define_elements()
        if not self.settings.disable_default_recorders:
            self.define_default_recorders()

    def opensees_define_loads(
        self,
        nodal_loads: dict[int, PointLoad],
        component_udl: dict[int, UDL],
        amplification_factor: float = 1.00,
        time_series_tag: int = 1,
        pattern_tag: int = 1,
    ) -> None:
        """
        Define loads in OpenSees.

        Defines a `Linear` `timeSeries` with a `Plain` `pattern` and
        assigns all loads in the `laodase`.

        Raises:
          TypeError: If the model type is invalid.
        """
        if time_series_tag not in self._time_series_tags:
            ops.timeSeries('Linear', time_series_tag)
            self._time_series_tags.append(time_series_tag)

        ops.pattern('Plain', pattern_tag, time_series_tag)

        # Point load on nodes
        for node_uid, point_load in nodal_loads.items():  # type: ignore
            ops.load(node_uid, *(v * amplification_factor for v in point_load))

        # UDL on components
        for component_uid, global_udl in component_udl.items():  # type: ignore
            component = self.model.components[component_uid]
            if component.tags & self.settings.ignore_by_tag:
                continue
            assert isinstance(component, BeamColumnAssembly)
            local_udls = component.calculate_element_udl(global_udl)
            for beamcolumn_element_uid, local_udl in local_udls.items():
                if self.model.dimensionality == '3D Frame':
                    ops.eleLoad(
                        '-ele',
                        beamcolumn_element_uid,
                        '-type',
                        '-beamUniform',
                        local_udl[1] * amplification_factor,
                        local_udl[2] * amplification_factor,
                        local_udl[0] * amplification_factor,
                    )
                elif self.model.dimensionality == '2D Frame':
                    ops.eleLoad(
                        '-ele',
                        beamcolumn_element_uid,
                        '-type',
                        '-beamUniform',
                        local_udl[1] * amplification_factor,
                        local_udl[0] * amplification_factor,
                    )
                else:
                    msg = f'Invalid model dimensionality: `{self.model.dimensionality}`.'
                    raise TypeError(msg)

    @staticmethod
    def opensees_define_mass(
        mass_registry: dict[int, PointMass], amplification_factor: float = 1.00
    ) -> None:
        """Define mass in OpenSees."""
        for node_uid, point_mass in mass_registry.items():  # type: ignore
            amplified_mass = [float(v) * amplification_factor for v in point_mass]
            if any(amplified_mass):  # Skip if all elements are zero
                ops.mass(node_uid, *amplified_mass)

    def define_default_recorders(self) -> None:
        """
        Create a set of default recorders.

        Does not define them in  OpenSees.

        Raises:
          ValueError: If the results directory is unspecified.
        """
        ndf = NDF[self.model.dimensionality]
        store_dir = self.settings.result_directory
        if store_dir is None:
            msg = 'Please specify a result directory in the analysis options.'
            raise ValueError(msg)
        node_recorder = NodeRecorder(
            uid_generator=self.model.uid_generator,
            file_name='node_displacements',
            recorder_type='Node',
            nodes=tuple(self.model.get_all_nodes().keys()),
            dofs=tuple(v + 1 for v in range(ndf)),
            response_type='disp',
            number_of_significant_digits=6,
            output_time=True,
        )
        self.recorders['default_node'] = node_recorder
        node_reaction_recorder = NodeRecorder(
            uid_generator=self.model.uid_generator,
            file_name='node_reactions',
            recorder_type='Node',
            nodes=tuple(self.model.get_all_nodes().keys()),
            dofs=tuple(v + 1 for v in range(ndf)),
            response_type='reaction',
            number_of_significant_digits=6,
            output_time=True,
        )
        self.recorders['default_node_reaction'] = node_reaction_recorder

        applicable_elements = []
        components = self.model.components.values()
        for component in components:
            if component.tags & self.settings.ignore_by_tag:
                continue
            elements = component.elements
            for element in elements.values():
                if isinstance(element, (ElasticBeamColumn, DispBeamColumn, Bar)):
                    if element.visibility.skip_opensees_definition:
                        continue
                    if (
                        isinstance(element, Bar)
                        and element.transf_type == 'Corotational'
                    ):
                        # Crurently `localForce` is not a valid
                        # recorder argument for corotational truss
                        # elements.
                        continue
                    applicable_elements.append(element.uid)

        element_force_recorder = ElementRecorder(
            uid_generator=self.model.uid_generator,
            file_name='basic_forces',
            recorder_type='Element',
            elements=tuple(applicable_elements),
            element_arguments=('localForce',),
            number_of_significant_digits=6,
            output_time=True,
        )
        self.recorders['default_basic_force'] = element_force_recorder

    def opensees_define_recorders(self) -> None:
        """Define recorders in OpenSees."""
        assert self.settings.result_directory is not None
        for recorder in self.recorders.values():
            # Update file name to include the base directory
            recorder.file_name = str(
                (Path(self.settings.result_directory) / recorder.file_name).resolve()
            )
            ops.recorder(*recorder.ops_args())

    def opensees_mck(self) -> None:
        """Get the mass, damping, and stiffness matrices."""
        ops.wipeAnalysis()
        ops.system('FullGeneral')
        ops.numberer(self.settings.numberer)
        ops.constraints(*self.settings.constraints)
        ops.analysis('Transient')
        # Mass
        ops.integrator('GimmeMCK', 1.0, 0.0, 0.0)
        ops.analyze(1, 0.0)
        # Number of equations in the model
        num = ops.systemSize()  # Has to be done after analyze
        # Convert to np array and reshape to NxN matrix
        mass_mat = np.array(ops.printA('-ret')).reshape((num, num))
        # Damping
        ops.integrator('GimmeMCK', 0.0, 1.0, 0.0)
        ops.analyze(1, 0.0)
        damping_mat = np.array(ops.printA('-ret')).reshape((num, num))
        # Stiffness
        ops.integrator('GimmeMCK', 0.0, 0.0, 1.0)
        ops.analyze(1, 0.0)
        stiffness_mat = np.array(ops.printA('-ret')).reshape((num, num))
        return {
            'number_of_equations': num,
            'stiffness_matrix': stiffness_mat,
            'damping_matrix': damping_mat,
            'mass_matrix': mass_mat,
        }

    def calculate_basic_forces(  # noqa: C901
        self,
        recorder_name: str,
        components: dict[int, ComponentAssembly],
        *,
        ndm: int,
        num_stations: int = 12,
    ) -> tuple[
        pd.DataFrame,  # Axial forces
        pd.DataFrame,  # Shear forces (Y)
        pd.DataFrame,  # Shear forces (Z)
        pd.DataFrame,  # Torsion
        pd.DataFrame,  # Bending moments (Y)
        pd.DataFrame,  # Bending moments (Z)
    ]:
        """
        Calculate basic forces at intermediate locations.

        This function calculates axial forces, shear forces (in Y and
        Z directions), torsion, and bending moments (in Y and Z
        directions) at multiple stations along each element based on
        the provided recorder data and element lengths.  The results
        are computed for either 2D or 3D elements and returned as a
        tuple of Pandas DataFrames.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame,
            pd.DataFrame, pd.DataFrame, pd.DataFrame]:
                A tuple containing six DataFrames in the following
                order:
                - Axial forces (`axial_df`)
                - Shear forces in the Y direction (`shear_y_df`)
                - Shear forces in the Z direction (`shear_z_df`) or
                  zeros for 2D
                - Torsion (`torsion_df`) or zeros for 2D
                - Bending moments in the Y direction (`moment_y_df`)
                  or zeros for 2D
                - Bending moments in the Z direction (`moment_z_df`)
                Each DataFrame is multi-indexed by 'element' and
                'station', with forces computed at evenly spaced
                stations along each element.

        Raises:
            ValueError: If the specified recorder is not found.
            TypeError: If the specified recorder is not an Element
              recorder.
            ValueError: If `ndf` (number of degrees of freedom) is not
              2 or 3.
            ValueError: If the recorder data do not have required
              column levels 'dof' or 'station'.
            ValueError: If element length information is missing for
              any element.
        """
        cache_key = (recorder_name, tuple(components.keys()), ndm, num_stations)

        # Check if the result is already cached
        if cache_key in self._basic_force_cache:
            return self._basic_force_cache[cache_key]

        recorder = self.recorders.get(recorder_name)
        if recorder is None:
            msg = f'Specified recorder not available: {recorder_name}.'
            raise ValueError(msg)
        if not isinstance(recorder, ElementRecorder):
            msg = f'The specified recorder (`{recorder_name}`) is not an Element recorder.'
            raise TypeError(msg)

        if isinstance(self, HasLoads):
            udls_global = self.load_registry.component_udl
        else:
            udls_global = {}

        udls_local = {}
        elements = {}
        for component_uid, component in components.items():
            if not isinstance(component, BeamColumnAssembly):
                if isinstance(component, BarAssembly):
                    elements.update(component.elements)
                continue
            for element in component.elements.values():
                if isinstance(element, (BeamColumnElement, Bar)):
                    elements[element.uid] = element
            elements.update(component.elements)
            component_global_udl = udls_global.get(component_uid)
            if component_global_udl:
                component_local_udls = component.calculate_element_udl(
                    component_global_udl
                )
                udls_local.update(component_local_udls)

        data = recorder.get_data()

        required_levels = {'dof', 'station'}
        if not required_levels.issubset(data.columns.names):
            msg = f'Data must have levels: {required_levels}'
            raise ValueError(msg)

        ndf = data.columns.get_level_values('dof').max()
        if ndf not in {3, 6}:
            msg = 'Must be either 2D or 3D Frame.'
            raise ValueError(msg)
        dof_data = {
            dof: data.xs(dof, level='dof', axis=1)
            for dof in data.columns.get_level_values('dof').unique()
        }
        dof_data_i = {
            dof: df.xs(0.0, level='station', axis=1) for dof, df in dof_data.items()
        }

        # dof_data_j = {
        #     dof: df.xs(1.0, level='station', axis=1) for dof, df in dof_data.items()
        # }

        missing_elements = [e for e in recorder.elements if e not in elements]
        if missing_elements:
            msg = f'Missing elements: {missing_elements}'
            raise ValueError(msg)

        element_lengths: dict[int, float] = {
            element.uid: element.clear_length()
            for element in elements.values()
            if isinstance(element, (BeamColumnElement, Bar))
        }

        locations = np.array(
            [
                np.linspace(0.00, element_lengths[element], num=num_stations)
                for element in recorder.elements
            ]
        )
        locations_expanded = locations[np.newaxis, :, :]

        columns = pd.MultiIndex.from_tuples(
            [  # noqa: C416
                (element, loc)
                for element, loc in product(
                    recorder.elements,
                    [
                        float(f'{v:.2f}')
                        for v in np.linspace(0.00, 1.00, num=num_stations)
                    ],
                )
            ],
            names=['element', 'station'],
        )

        # ~~~ Local X axis axial force ~~~

        data_i_axial = dof_data_i[1].to_numpy()[:, :, np.newaxis]
        # data_j_axial = dof_data_j[1]
        w_data_axial = np.array(
            [
                udls_local[element_uid][0]
                if udls_local.get(element_uid) is not None
                else 0.00
                for element_uid in recorder.elements
            ]
        )[np.newaxis, :, np.newaxis]

        result_axial = -(data_i_axial + w_data_axial * locations_expanded)
        result_axial_flattened = result_axial.reshape(result_axial.shape[0], -1)
        axial_df = pd.DataFrame(
            result_axial_flattened, index=dof_data_i[1].index, columns=columns
        )

        # # check: results at station '1.00' should be equal to `data_j`.
        # pd.testing.assert_frame_equal(
        #     axial_df.xs(1.0, level='station', axis=1),
        #     data_j_axial,
        #     check_exact=False,
        #     atol=1e-6,
        # )

        # ~~~ Local Y axis shear force ~~~

        data_i_shear_y = dof_data_i[2].to_numpy()[:, :, np.newaxis]
        # data_j_shear_y = dof_data_j[2]
        w_data_shear_y = np.array(
            [
                udls_local[element_uid][1]
                if udls_local.get(element_uid) is not None
                else 0.00
                for element_uid in recorder.elements
            ]
        )[np.newaxis, :, np.newaxis]

        result_shear_y = data_i_shear_y + w_data_shear_y * locations_expanded
        result_shear_y_flattened = result_shear_y.reshape(
            result_shear_y.shape[0], -1
        )
        shear_y_df = pd.DataFrame(
            result_shear_y_flattened, index=dof_data_i[1].index, columns=columns
        )

        # # check: results at station '1.00' should be equal to `data_j`.
        # pd.testing.assert_frame_equal(
        #     shear_y_df.xs(1.0, level='station', axis=1),
        #     - data_j_shear_y,
        #     check_exact=False,
        #     atol=1e-6,
        # )

        # ~~~ Local Z axis shear force ~~~

        # In 2D this is zero.
        if ndm == TWO_DIMENSIONAL:
            shear_z_df = pd.DataFrame(
                0.00, index=axial_df.index, columns=shear_y_df.columns
            )
        elif ndm == THREE_DIMENSIONAL:
            data_i_shear_z = dof_data_i[3].to_numpy()[:, :, np.newaxis]
            # data_j_shear_z = dof_data_j[3]
            w_data_shear_z = np.array(
                [
                    udls_local[element_uid][2]
                    if udls_local.get(element_uid) is not None
                    else 0.00
                    for element_uid in recorder.elements
                ]
            )[np.newaxis, :, np.newaxis]

            result_shear_z = data_i_shear_z + w_data_shear_z * locations_expanded
            result_shear_z_flattened = result_shear_z.reshape(
                result_shear_z.shape[0], -1
            )
            shear_z_df = pd.DataFrame(
                result_shear_z_flattened, index=dof_data_i[1].index, columns=columns
            )

            # # check: results at station '1.00' should be equal to `data_j`.
            # pd.testing.assert_frame_equal(
            #     shear_z_df.xs(1.0, level='station', axis=1),
            #     data_j_shear_z,
            #     check_exact=False,
            #     atol=1e-6,
            # )

        # ~~~ Local X axis torsional moment ~~~

        # In 2D this is zero.
        if ndm == TWO_DIMENSIONAL:
            torsion_df = pd.DataFrame(
                0.00, index=axial_df.index, columns=shear_y_df.columns
            )
        elif ndm == THREE_DIMENSIONAL:
            data_i_torsion = dof_data_i[4].to_numpy()[:, :, np.newaxis]
            # data_j_torsion = dof_data_j[4]

            result_torsion = -data_i_torsion * np.ones_like(locations_expanded)
            result_torsion_flattened = result_torsion.reshape(
                result_torsion.shape[0], -1
            )
            torsion_df = pd.DataFrame(
                result_torsion_flattened, index=dof_data_i[1].index, columns=columns
            )

            # # check: results at station '1.00' should be equal to `data_j`.
            # pd.testing.assert_frame_equal(
            #     torsion_df.xs(1.0, level='station', axis=1),
            #     data_j_torsion,
            #     check_exact=False,
            #     atol=1e-6,
            # )

        # ~~~ Local Y axis bending moment (typically the weak axis) ~~~
        if ndm == TWO_DIMENSIONAL:
            moment_y_df = pd.DataFrame(
                0.00, index=axial_df.index, columns=shear_y_df.columns
            )
        elif ndm == THREE_DIMENSIONAL:
            data_i_moment_y = dof_data_i[5].to_numpy()[:, :, np.newaxis]
            # data_j_moment_y = dof_data_j[5]

            # Already obtained: {w_data_shear_z, data_i_shear_z, data_j_shear_z}

            result_moment_y = (
                locations_expanded**2 * 0.50 * w_data_shear_z
                + locations_expanded * data_i_shear_z
                + data_i_moment_y
            )
            result_moment_y_flattened = result_moment_y.reshape(
                result_moment_y.shape[0], -1
            )
            moment_y_df = pd.DataFrame(
                result_moment_y_flattened, index=dof_data_i[1].index, columns=columns
            )

            # # check: results at station '1.00' should be equal to `data_j`.
            # pd.testing.assert_frame_equal(
            #     moment_y_df.xs(1.0, level='station', axis=1),
            #     data_j_moment_y,
            #     check_exact=False,
            #     atol=1e-6,
            # )

        # ~~~ Local Z axis bending moment (typically the strong axis) ~~~
        if ndm == TWO_DIMENSIONAL:
            data_i_moment_z = dof_data_i[3].to_numpy()[:, :, np.newaxis]
            # data_j_moment_z = dof_data_j[3]

        elif ndm == THREE_DIMENSIONAL:
            data_i_moment_z = dof_data_i[6].to_numpy()[:, :, np.newaxis]
            # data_j_moment_z = dof_data_j[6]

        # Already obtained: {w_data_shear_y, data_i_shear_y, data_j_shear_y}

        result_moment_z = (
            locations_expanded**2 * 0.50 * w_data_shear_y
            + locations_expanded * data_i_shear_y
            - data_i_moment_z
        )
        result_moment_z_flattened = result_moment_z.reshape(
            result_moment_z.shape[0], -1
        )
        moment_z_df = pd.DataFrame(
            result_moment_z_flattened, index=dof_data_i[1].index, columns=columns
        )

        # # check: results at station '1.00' should be equal to `data_j`.
        # pd.testing.assert_frame_equal(
        #     moment_z_df.xs(1.0, level='station', axis=1),
        #     data_j_moment_z,
        #     check_exact=False,
        #     atol=1e-6,
        # )

        result = (
            axial_df,
            shear_y_df,
            shear_z_df,
            torsion_df,
            moment_y_df,
            moment_z_df,
        )
        self._basic_force_cache[cache_key] = result

        return result

    def run_static(
        self, load_case: StaticLoadCase, num_steps: int = 1, *, wipe: bool = True
    ) -> None:
        """Run a static analysis."""
        self.initialize_logger()
        self.log('Running a static analysis.')
        self.log('Defining model in OpenSees.')
        self.opensees_define_model()
        self.opensees_define_recorders()
        self.opensees_define_node_restraints(
            load_case.fixed_supports, load_case.elastic_supports
        )
        self.opensees_define_node_constraints(load_case.rigid_diaphragm)
        self.log('Defining loads in OpenSees.')
        self.opensees_define_loads(
            load_case.load_registry.nodal_loads,
            load_case.load_registry.component_udl,
        )
        self.log('Setting up analysis.')
        self.log(f'Setting system solver to {self.settings.system}')
        ops.system(self.settings.system)
        self.log(f'Setting numberer to {self.settings.numberer}')
        ops.numberer(self.settings.numberer)
        self.log(f'Setting constraints to {self.settings.constraints}')
        ops.constraints(*self.settings.constraints)
        self.log("Setting test to ('EnergyIncr', 1.0e-8, 20, 3)")
        ops.test('EnergyIncr', 1.0e-8, 20, 3)
        self.log('Setting algorithm to KrylovNewton')
        ops.algorithm('KrylovNewton')
        self.log(f'Setting integrator to LoadControl with {num_steps} steps')
        ops.integrator('LoadControl', 1.00 / num_steps)
        self.log('Setting analysis to Static')
        ops.analysis('Static')
        self.log('Analyzing.')
        out = ops.analyze(num_steps)
        assert out == 0, 'Analysis failed.'
        if wipe:
            ops.wipe()
        self.log('Analysis finished.')

    def run_pushover(
        self,
        target_displacements: list[float | None],
        control_node_uid: int,
        dof: int,
        displ_incr: float,
        *,
        pattern_tag: int,
        max_steps: int = 50,
        norm: float = 1e-8,
    ) -> None:
        """
        Run a pushover analysis.

        Raises:
          ValueError: If it fails to unload.
        """
        curr_displ = ops.nodeDisp(control_node_uid, dof)

        self.log('Starting pushover analysis')
        ops.system(self.settings.system)
        ops.numberer(self.settings.numberer)
        ops.constraints(*self.settings.constraints)

        total_fail = False
        num_subdiv = 0
        num_times = 0
        algorithm_idx = 0

        scale = [1.0, 1.0e-1, 1.0e-2, 1.0e-3]
        algorithms = [('KrylovNewton',), ('KrylovNewton', 'initial')]

        for i_loop, target_displacement in enumerate(target_displacements):
            if total_fail:
                break
            if target_displacement is not None:
                # determine push direction
                if curr_displ < target_displacement:
                    displ_incr = abs(displ_incr)
                    sign = +1.00
                else:
                    displ_incr = -abs(displ_incr)
                    sign = -1.00

                while curr_displ * sign < target_displacement * sign:
                    # determine increment
                    if (
                        abs(curr_displ - target_displacement)
                        < abs(displ_incr) * scale[num_subdiv]
                    ):
                        incr = sign * abs(curr_displ - target_displacement)
                    else:
                        incr = displ_incr * scale[num_subdiv]

                    ops.test('NormDispIncr', norm, max_steps, 0)
                    ops.algorithm(*algorithms[algorithm_idx])
                    ops.system(self.settings.solver)
                    ops.integrator(
                        'DisplacementControl', control_node_uid, dof, incr
                    )
                    ops.analysis('Static')
                    flag = ops.analyze(1)
                    if flag != 0:
                        if num_subdiv == len(scale) - 1:
                            # can't refine further
                            print('Analysis failed to converge')
                            self.log(f'Analysis failed at disp {curr_displ:.5f}')
                            total_fail = True
                            break
                        # can still reduce step size
                        if algorithm_idx != len(algorithms) - 1:
                            algorithm_idx += 1
                        else:
                            algorithm_idx = 0
                            num_subdiv += 1
                            # how many times to run with reduced step size
                            num_times = 50
                    else:
                        # analysis was successful
                        if num_times != 0:
                            num_times -= 1

                        curr_displ = ops.nodeDisp(control_node_uid, dof)
                        print(
                            f'Loop ({i_loop + 1}/'
                            f'{len(target_displacements)}) | '
                            'Target displacement: '
                            f'{target_displacement:.2f}'
                            f' | Current: {curr_displ:.4f}',
                            end='\r',
                        )
                        algorithm_idx = 0
                        if num_subdiv != 0:
                            if num_times == 0:
                                num_subdiv -= 1
                                num_times = 10

            else:
                # Need to unload
                ops.test('NormDispIncr', norm[num_subdiv], max_steps, 0)
                ops.algorithm(*algorithms[algorithm_idx])
                current_load = ops.getLoadFactor(pattern_tag)
                load_threshold = 1e-4
                while current_load > load_threshold:
                    increment = -current_load / 10.00
                    ops.integrator('LoadControl', increment)
                    flag = ops.analyze(1)
                    if flag != 0:
                        msg = 'Failed to unload.'
                        raise ValueError(msg)
                    current_load = ops.getLoadFactor(pattern_tag)
                    curr_displ = ops.nodeDisp(control_node_uid, dof)
                    print(
                        f'Loop ({i_loop + 1}/'
                        f'{len(target_displacements)}) | '
                        'Target displacement: '
                        f'(Unloading)'
                        f' | Current: {curr_displ:.4f}',
                        end='\r',
                    )

        print('Analysis finished.')

    @staticmethod
    def opensees_define_transient_time_series_and_pattern(
        direction: int,
        ag_vec: np.ndarray,
        dt: float,
        factor: float,
        time_series_tag: int,
        pattern_tag: int,
    ) -> None:
        """Define a time series and pattern for a transient analysis."""
        ops.timeSeries(
            'Path',
            time_series_tag,
            '-dt',
            dt,
            '-values',
            *ag_vec,
            '-factor',
            factor,
        )
        ops.pattern(
            'UniformExcitation', pattern_tag, direction, '-accel', time_series_tag
        )

    @staticmethod
    def opensees_define_viscous_damping_rayleigh(
        period_1: float, period_2: float, ratio: float
    ) -> None:
        """
        Assign Rayleigh damping.

        Assigns Rayleigh damping given two periods and the damping
        ratio.
        """
        w_i = 2 * np.pi / period_1
        w_j = 2 * np.pi / period_2
        zeta_i = ratio
        zeta_j = ratio
        a_mat = np.array([[1 / w_i, w_i], [1 / w_j, w_j]])
        b_vec = np.array([zeta_i, zeta_j])
        x_sol = np.linalg.solve(a_mat, 2 * b_vec)
        ops.rayleigh(x_sol[0], 0.0, 0.0, x_sol[1])
        # https://portwooddigital.com/2020/11/08/rayleigh-damping-coefficients/
        # --thanks, prof. Scott

    @staticmethod
    def opensees_define_viscous_damping_stiffness(
        period: float, ratio: float
    ) -> None:
        """Assign stiffness-proportional damping."""
        ops.rayleigh(0.00, 0.0, 0.0, ratio * period / np.pi)

    @staticmethod
    def opensees_define_viscous_damping_modal(num_modes: int, ratio: float) -> None:
        """Assign modal damping."""
        ops.eigen(num_modes)
        ops.modalDampingQ(ratio)

    @staticmethod
    def opensees_define_viscious_damping_modal_and_stiffness(
        ratio_stiffness: float, period: float, ratio_modal: float, num_modes: int
    ) -> None:
        """
        Assign modal + stiffness-proportional damping.

        Assigns modal combined with stiffness proportional damping.
        """
        alpha_1 = ratio_stiffness * period / np.pi
        ops.rayleigh(
            0.00,
            0.0,
            0.0,
            alpha_1,
        )
        omega_squareds = np.array(ops.eigen(num_modes))
        damping_vals = ratio_modal - alpha_1 * np.sqrt(omega_squareds) / 2.00
        assert np.min(damping_vals) > 0.00
        ops.modalDampingQ(*damping_vals)

    def run_transient(
        self,
        *,
        analysis_time_increment: float,
        finish_time: float,
        print_progress: bool = True,
        time_limit: float | None = None,
        transient_drift_check_setup: TransientDriftCheckSetup | None = None,
    ) -> None:
        """
        Run a transient analysis.

        Arguments:
          analysis_time_increment: Time increment.
          finish_time: Specify a target time (s) to stop the analysis.
          print_progress: Controls whether the current time is printed out.
          time_limit: Maximum analysis time allowed, in hours.
          When reached, the analysis is interrupted.

        Raises:
          ValueError: If no mass is assigned.
        """
        # Check if there is any mass
        total_mass = 0.00
        node_tags = ops.getNodeTags()
        ndf = NDF[self.model.dimensionality]
        for node in node_tags:
            total_mass += np.sum([ops.nodeMass(node, i + 1) for i in range(ndf)])
        if total_mass == 0.00:
            msg = 'No mass!'
            raise ValueError(msg)

        # Constants
        tolerance: float = 1e-9
        max_iterations: int = 100
        time_limit_seconds: float | None = (
            time_limit * 3600.0 if time_limit else None
        )

        self.log('Starting transient analysis')

        # Set up OpenSees parameters
        ops.numberer(self.settings.numberer)
        ops.constraints(*self.settings.constraints)
        ops.system(self.settings.solver)
        ops.test('EnergyIncr', tolerance, max_iterations, 0)
        ops.integrator('Newmark', 0.5, 0.25)
        ops.algorithm('Newton')
        ops.analysis('Transient')

        # Progress bar
        pbar = self._initialize_progress_bar(
            print_progress=print_progress, finish_time=finish_time
        )

        start_time: float = perf_counter()
        last_log_time: float = start_time
        curr_time: float = 0.0

        self.log('Initiating time traversal')
        self.log(f'Max time step: {analysis_time_increment}')

        try:
            self._perform_analysis_loop(
                curr_time,
                finish_time,
                analysis_time_increment,
                start_time,
                last_log_time,
                pbar,
                time_limit_seconds,
                transient_drift_check_setup,
            )
        except KeyboardInterrupt:
            self.log('Analysis interrupted')
            self._logger.warning('Analysis interrupted')

        self.log('Analysis finished.')
        if pbar is not None:
            pbar.close()

    def run_transient_dampen_residual_response(
        self,
        prior_load_patterns: list,
        analysis_time_increment: float,
        *,
        print_norm: bool = True,
        velocity_threshold: float = 1e-2,
        rayleigh_factor: float = 1.00,
        absolute_tolerance: float = 1e-4,
    ) -> None:
        """Dampen out any residual motion."""
        for load_pattern in prior_load_patterns:
            ops.remove('loadPattern', load_pattern)

        ops.rayleigh(rayleigh_factor, 0.00, 0.0, 0.00)

        vel_norm = np.inf
        while vel_norm > velocity_threshold:
            check = ops.analyze(1, analysis_time_increment)

            if check != 0:
                self.log('Failed to dampen out residual motion: Analysis fails.')
                break

            vel = np.zeros(6)
            node_tags = ops.getNodeTags()
            num_nodes = len(node_tags)
            for ntag in node_tags:
                vel += [x / num_nodes for x in ops.nodeVel(ntag)]
            previous_norm = vel_norm
            vel_norm = np.sqrt(vel @ vel)

            if print_norm:
                print(f'{vel_norm:5.3e} > {velocity_threshold:5.3e}', end='\r')

            if np.abs(vel_norm - previous_norm) < absolute_tolerance:
                self.log(
                    'Failed to dampen out residual motion: '
                    'Velocity norm stopped decreasing.'
                )
                break

        print()

    @staticmethod
    def _initialize_progress_bar(
        *, print_progress: bool, finish_time: float
    ) -> tqdm.std.tqdm | None:
        """Initialize a progress bar if requested."""
        if not print_progress:
            return None

        pbar: tqdm.std.tqdm = tqdm(
            total=finish_time,
            bar_format='{percentage:3.0f}%|{bar:25}| [{elapsed}<{remaining}, {rate_fmt}{postfix}]',
        )
        return pbar

    def _perform_analysis_loop(
        self,
        curr_time: float,
        finish_time: float,
        analysis_time_increment: float,
        start_time: float,
        last_log_time: float,
        pbar: tqdm.std.tqdm | None,
        time_limit_seconds: float | None,
        transient_drift_check_setup: TransientDriftCheckSetup | None,
    ) -> None:
        """Perform the main loop of a transient analysis."""
        scale: tuple[float, float, float, float] = (1.0, 1.0e-1, 1.0e-2, 1.0e-3)
        algorithms: tuple[tuple[str], tuple[str, str, str], tuple[str]] = (
            ('KrylovNewton',),
            ('KrylovNewton', 'initial', 'initial'),
            ('NewtonLineSearch',),
        )
        num_subdiv: int = 0
        num_times: int = 0
        algorithm_idx: int = 0
        analysis_failed: bool = False

        while curr_time + common.EPSILON < finish_time:
            if analysis_failed:
                break

            ops.test('EnergyIncr', 1e-9, 100, 0)
            ops.algorithm(*algorithms[algorithm_idx])
            check: int = ops.analyze(1, analysis_time_increment * scale[num_subdiv])

            if check != 0:
                num_subdiv, num_times, algorithm_idx, should_stop = (
                    self._handle_analysis_failure(
                        curr_time,
                        num_subdiv,
                        algorithm_idx,
                        num_times,
                        scale,
                        algorithms,
                    )
                )
                if should_stop:
                    break
            else:
                curr_time, last_log_time, should_stop = (
                    self._handle_successful_analysis(
                        curr_time,
                        finish_time,
                        pbar,
                        start_time,
                        last_log_time,
                        num_subdiv,
                        num_times,
                        time_limit_seconds,
                        transient_drift_check_setup,
                    )
                )
                if should_stop:
                    break  # Stop the loop if the time limit is reached

    def _handle_analysis_failure(
        self,
        curr_time: float,
        num_subdiv: int,
        algorithm_idx: int,
        num_times: int,
        scale: tuple[float, float, float, float],
        algorithms: tuple[tuple[str], tuple[str, str, str], tuple[str]],
    ) -> bool:
        """Handle an analysis failure by adjusting parameters or logging failure."""
        if num_subdiv == len(scale) - 1:
            self.log('Analysis failed to converge.')
            self._logger.warning(
                f'Analysis failed at time {curr_time:.5f} and cannot continue.'
            )
            return num_subdiv, num_times, algorithm_idx, True  # Analysis failed

        if algorithm_idx < len(algorithms) - 1:
            return (
                num_subdiv,
                num_times,
                algorithm_idx,
                False,
            )  # Try the next algorithm

        # Increase subdivisions if all algorithms have been attempted
        num_subdiv += 1
        num_times = 50  # Reset retries
        algorithm_idx = 0

        return num_subdiv, num_times, algorithm_idx, False

    def _handle_successful_analysis(
        self,
        curr_time: float,
        finish_time: float,
        pbar: tqdm.std.tqdm | None,
        start_time: float,
        last_log_time: float,
        num_subdiv: int,
        num_times: int,
        time_limit_seconds: float | None,
        transient_drift_check_setup: TransientDriftCheckSetup | None,
    ) -> tuple[float, float, bool]:
        """Handle successful analysis updates."""
        prev_time: float = curr_time
        curr_time = float(ops.getTime())
        test_iter: int = ops.testIter()

        # Update progress bar
        if pbar is not None:
            pbar.set_postfix(
                {'time': f'{curr_time:.4f}/{finish_time:.2f} [{test_iter}]'}
            )
            pbar.update(curr_time - prev_time)

        # Periodic logging
        if perf_counter() - last_log_time > 5.00 * 60.00:  # 5 min
            last_log_time = perf_counter()
            running_time: float = last_log_time - start_time
            remaining_time: float = finish_time - curr_time
            average_speed: float = curr_time / running_time
            est_remaining_dur: float = remaining_time / average_speed
            self.log(
                f'Analysis status: {{curr: {curr_time:.2f}, target: {finish_time:.2f}, '
                f'num_subdiv: {num_subdiv}, ~ {est_remaining_dur:.0f} s to finish}}'
            )

        # Time limit check
        if time_limit_seconds and (perf_counter() - start_time) > time_limit_seconds:
            self._logger.warning(
                f'Analysis interrupted at time {curr_time:.5f} because the time limit was reached.'
            )
            return curr_time, last_log_time, True  # Return flag to stop the loop

        # Maximum drift check
        if transient_drift_check_setup is not None:
            node_uids = transient_drift_check_setup.node_uids
            drift_limit = transient_drift_check_setup.drift_limit
            elevation_dof = transient_drift_check_setup.elevation_dof
            # We need at least two nodes
            min_nodes_required = 2
            assert len(node_uids) > min_nodes_required
            assert elevation_dof in {1, 2, 3}
            if elevation_dof == 1:
                other_dofs = (2, 3)
            elif elevation_dof == 2:  # noqa: PLR2004
                other_dofs = (1, 3)
            else:
                other_dofs = (1, 2)
            node_pairs = list(zip(node_uids, node_uids[1:]))
            for bottom_node, top_node in node_pairs:
                bottom_elev = ops.nodeCoord(bottom_node)[elevation_dof - 1]
                top_elev = ops.nodeCoord(top_node)[elevation_dof - 1]
                for other_dof in other_dofs:
                    top_disp = ops.nodeDisp(top_node, other_dof)
                    bottom_disp = ops.nodeDisp(bottom_node, other_dof)
                    drift = (top_disp - bottom_disp) / (top_elev - bottom_elev)
                    if drift > drift_limit:
                        self.log(
                            f'Drift limit reached: {drift * 100:.2}% at dofs ({top_node}, {bottom_node}). '
                            f'Time: {curr_time:.3f} s.'
                            f'Stopping analysis.'
                        )
                        return curr_time, last_log_time, True  # Stop

        if num_times != 0:
            num_times -= 1
        elif num_subdiv != 0:
            num_subdiv -= 1
            num_times = 50

        return curr_time, last_log_time, False  # Continue analysis


@dataclass()
class ModalAnalysisSettings(AnalysisSettings):
    """Modal analysis settings."""

    num_modes: int = field(default=3)
    retrieve_basic_forces: bool = field(default=True)


@dataclass(repr=False)
class ModalAnalysis(Analysis):
    """Modal analysis."""

    load_case: ModalLoadCase = field(default_factory=ModalLoadCase)
    settings: ModalAnalysisSettings = field(default_factory=ModalAnalysisSettings)
    periods: list[float] = field(default_factory=list)
    _is_executed: bool = field(default=False, init=False)

    def run_modal(self) -> None:  # noqa: C901  # type: ignore
        """
        Run the modal analysis.

        Raises:
          ValueError: If a solver known to not work with `eigen` is
           specified.
        """
        self.initialize_logger()
        self.log('Running a modal analysis.')

        ndf = NDF[self.model.dimensionality]

        if self.settings.disable_default_recorders is True:
            self.settings.disable_default_recorders = False
            msg = (
                'Default recorders are required for '
                'modal analysis and were force-enabled.'
            )
            self.log(msg)
            print(msg)
            # TODO(JVM): turn into a warning.

        self.log('Defining model in OpenSees.')
        self.opensees_define_model()
        self.opensees_define_recorders()
        self.opensees_define_node_restraints(
            self.load_case.fixed_supports, self.load_case.elastic_supports
        )
        self.opensees_define_node_constraints(self.load_case.rigid_diaphragm)
        self.log('Defining mass in OpenSees.')
        self.opensees_define_mass(self.load_case.mass_registry)

        self.log('Setting up analysis.')
        if self.settings.system.lower() in {'sparsesym', 'sparsespd'}:
            msg = (
                f'{self.settings.solver} is unable '
                'to run a modal analysis. Use UmfPack.'
            )
            self.log(msg)
            raise ValueError(msg)

        ops.system(self.settings.system)
        ops.numberer(self.settings.numberer)
        ops.constraints(*self.settings.constraints)

        self.log('Analyzing.')
        lambda_values = np.array(ops.eigen(self.settings.num_modes))

        self.log('Calculating periods.')
        omega_values = np.sqrt(lambda_values)
        periods = 2.0 * np.pi / omega_values
        self.periods = periods
        self.log(f'Periods: {periods}')

        self.log('Retrieving node eigenvectors.')

        node_recorder = self.recorders['default_node']

        assert isinstance(node_recorder, NodeRecorder)
        nodes = node_recorder.nodes
        data = {}
        for mode in range(1, self.settings.num_modes + 1):
            for node in nodes:
                disp = ops.nodeEigenvector(node, mode)
                data[node, mode] = disp
        eigenvectors = pd.DataFrame(
            data.values(),
            index=pd.MultiIndex.from_tuples(data.keys(), names=['node', 'mode']),
            columns=range(1, ndf + 1),
        )
        eigenvectors.columns.name = 'dof'
        eigenvectors = eigenvectors.unstack(level='node')  # noqa: PD010
        eigenvectors.columns = eigenvectors.columns.reorder_levels(['node', 'dof'])
        node_order = eigenvectors.columns.get_level_values('node').unique()
        eigenvectors = eigenvectors.loc[:, node_order]
        self.recorders['default_node'].set_data(eigenvectors)
        self.log('Retrieved node eigenvectors.')

        basic_force_data = {}

        if self.settings.retrieve_basic_forces:
            self.log('Obtaining basic forces for each mode.')
            # Recover basic forces
            self.log('   Wiping OpenSees domain.')
            ops.wipe()
            progress_bar = tqdm(
                range(1, self.settings.num_modes + 1),
                ncols=80,
                desc='Processing modes',
                unit='mode',
                position=0,
                leave=False,
            )
            for mode in range(1, self.settings.num_modes + 1):
                self.log(f'  Working on mode {mode}.')
                self.log('  Defining model in OpenSees.')
                self.opensees_define_model()
                self.opensees_define_recorders()
                self.opensees_define_node_restraints(
                    self.load_case.fixed_supports, self.load_case.elastic_supports
                )
                self.opensees_define_node_constraints(self.load_case.rigid_diaphragm)
                mode_eigenvectors = eigenvectors.loc[mode, :].copy()
                # ignore the node-dof pairs with a fixed constraint.
                to_drop = []
                for uid, support in self.load_case.fixed_supports.items():
                    for i, dof in enumerate(support):
                        if dof:
                            to_drop.append((uid, i + 1))
                mode_eigenvectors = mode_eigenvectors.drop(to_drop)
                if (
                    bool(
                        (
                            mode_eigenvectors.isna() | np.isinf(mode_eigenvectors)
                        ).any()
                    )
                    is True
                ):
                    self.log(
                        f'  NaNs or infs present in displacement data, skipping mode {mode}.'
                    )
                    continue

                self.log('  Setting up Sp constraints.')
                ops.timeSeries('Constant', 1)
                ops.pattern('Plain', 1, 1)

                for key, displacement in mode_eigenvectors.items():
                    node, dof = key
                    ops.sp(node, dof, displacement)

                self.log('  Setting up analysis.')
                ops.integrator('LoadControl', 0.0)
                ops.constraints('Transformation')
                ops.algorithm('Linear')
                ops.numberer(self.settings.numberer)
                ops.system(self.settings.system)
                ops.analysis('Static')

                self.log('  Analyzing.')
                out = ops.analyze(1)
                assert out == 0, 'Analysis failed.'
                # The recorders should have captured the results here.
                self.log('  Retrieving data from recorder output.')
                # TODO(JVM): read basic_force_data from the recorder, add
                # in the dict.

                # Verifying displacements are correct.
                for key, value in mode_eigenvectors.items():
                    node_uid, dof = key
                    assert np.allclose(ops.nodeDisp(node_uid, dof), value)

                self.log('   Wiping OpenSees domain.')
                # Doing this before reading the basic force recorder data
                # ensures the recorder's buffer will have been flushed.
                ops.wipe()

                basic_force_data[mode] = self.recorders[
                    'default_basic_force'
                ].get_data()
                basic_force_data[mode].index.name = 'mode'
                basic_force_data[mode].index = [mode]

                progress_bar.update(1)

            progress_bar.close()
            self.log('Obtained basic forces for each mode.')
            self.recorders['default_basic_force'].set_data(
                pd.concat(basic_force_data.values(), axis=0)
            )

        self.log('Analysis finished.')
        ops.wipe()
        self._is_executed = True


def ensure_minmax_level_exists_or_add(data: pd.DataFrame) -> pd.DataFrame:
    """
    Add a 'min/max' column level if it doesn't exist.

    Assigns everything to 'max', and duplicates columns for 'min'.

    Args:
        data: Input DataFrame with MultiIndex columns.

    Returns:
        Updated DataFrame with 'min/max' as the outermost column
        level.
    """
    columns = data.columns

    if 'min/max' not in columns.names:
        new_columns = pd.MultiIndex.from_tuples(
            [(*col, minmax) for col in columns for minmax in ('max', 'min')],
            names=[*list(columns.names), 'min/max'],
        )
        # Repeat the data for 'max' and 'min'
        repeated_data = pd.concat([data, data], axis=1)
        repeated_data.columns = new_columns
        return repeated_data

    return data


def combine_single(
    df1: pd.DataFrame, df2: pd.DataFrame, action: Literal['add', 'envelope']
) -> pd.DataFrame:
    """
    Combine two DataFrames based on the specified action.

    Args:
        df1: First DataFrame.
        df2: Second DataFrame.
        action: Action to perform:
            - 'add': Element-wise addition of the DataFrames. Missing values are treated as 0.
            - 'envelope': Take the largest of the maxes and the
              smallest of the mins.

    Returns:
        Combined DataFrame based on the action.

    Raises:
      ValueError: If an unknown action is specified.
    """
    # Reindex to ensure alignment and fill missing data with 0
    df1_aligned = df1.reindex(columns=df2.columns.union(df1.columns), fill_value=0)
    df2_aligned = df2.reindex(columns=df2.columns.union(df1.columns), fill_value=0)

    if action == 'add':
        combined = df1_aligned + df2_aligned
    elif action == 'envelope':
        df1 = ensure_minmax_level_exists_or_add(df1)
        df2 = ensure_minmax_level_exists_or_add(df2)
        if not np.all(
            df1.columns.names == df2.columns.names
        ) or not df1.columns.equals(df2.columns):
            msg = 'Cannot align DataFrames with different columns'
            raise ValueError(msg)
        max_df = pd.DataFrame(
            np.maximum(
                df1.xs('max', level='min/max', axis=1),
                df2.xs('max', level='min/max', axis=1),
            ),
            index=df1.index,
            columns=df1.xs('max', level='min/max', axis=1).columns,
        )
        min_df = pd.DataFrame(
            np.minimum(
                df1.xs('min', level='min/max', axis=1),
                df2.xs('min', level='min/max', axis=1),
            ),
            index=df1.index,
            columns=df1.xs('min', level='min/max', axis=1).columns,
        )
        combined = pd.concat([max_df, min_df], axis=1)
        combined.columns = pd.MultiIndex.from_product(
            [max_df.columns.levels[0], max_df.columns.levels[1], ['max', 'min']],
            names=[*max_df.columns.names, 'min/max'],
        )

    else:
        msg = 'Action must be one of `add` or `envelope`.'
        raise ValueError(msg)

    return combined


def combine(
    dfs: list[pd.DataFrame], action: Literal['add', 'envelope']
) -> pd.DataFrame:
    """
    Combine multiple DataFrames sequentially based on the specified action.

    Args:
        dfs: List of DataFrames to combine.
        action: Action to perform:
            - 'add': Element-wise addition of the DataFrames.
            - 'envelope': Take the largest of the maxes and the
              smallest of the mins.

    Returns:
        Combined DataFrame based on the action.

    Raises:
        ValueError: If less than two DataFrames are provided.
    """
    min_df_count = 2
    if len(dfs) < min_df_count:
        msg = 'At least two DataFrames are required to combine.'
        raise ValueError(msg)

    # Combine DataFrames sequentially
    combined_df = dfs[0]
    for df in dfs[1:]:
        combined_df = combine_single(combined_df, df, action)

    return combined_df


@dataclass(repr=False)
class AnalysisRegistry:
    """Analysis registry."""

    load_case_registry: LoadCaseRegistry
    analysis_objects: dict[Analysis] = field(default_factory=dict)

    def run_static_batch(self) -> None:
        """Run a batch of static analyses."""
        # Determine the base directory for results
        base_dir = (
            Path(self.load_case_registry.result_setup.directory)
            if self.load_case_registry.result_setup.directory
            else Path(tempfile.mkdtemp())
        )
        base_dir.mkdir(parents=True, exist_ok=True)
        self.load_case_registry.result_setup.directory = str(base_dir.resolve())

        cases_dict = self.load_case_registry.get_load_cases()
        num_cases = len(cases_dict)
        progress_bar = tqdm(
            total=num_cases,
            ncols=80,
            desc='Processing cases',
            unit='case',
            leave=False,
        )
        for load_case_name, load_case in cases_dict.items():
            case_type = load_case.get_load_case_type()
            progress_bar.set_description(f'Processing {case_type}: {load_case_name}')
            # Create a subdirectory for each load case
            case_dir = base_dir / f'{case_type}_{load_case_name}'
            case_dir.mkdir(parents=True, exist_ok=True)
            analysis = Analysis(
                self.load_case_registry.model,
                AnalysisSettings(num_steps=1, result_directory=str(case_dir)),
            )
            analysis.run_static(self.load_case)
            self.analysis_objects[load_case_name] = analysis
            progress_bar.update(1)
        progress_bar.close()

    def get_combined_results(
        self, recorder_name: str, combination: dict[str, float]
    ) -> pd.DataFrame:
        """
        Get results for a specific load combination.

        Raises:
          ValueError: If the recorder is not found for some load case.

        Returns:
          The results.
        """
        associated_load_cases = set(combination.keys())
        load_case_objects = {
            load_case: self.load_case_registry.find_load_case_by_name(load_case)
            for load_case in associated_load_cases
        }
        all_data = []
        for load_case_name, scale_factor in combination.items():
            load_case = load_case_objects[load_case_name]
            if recorder_name not in self.analysis_objects[load_case_name].recorders:
                msg = (
                    f'Recorder not found: {recorder_name} '
                    f'for load case {load_case}.'
                )
                raise ValueError(msg)
            data = (
                self.analysis_objects[load_case_name]
                .recorders[recorder_name]
                .get_data()
            )
            if isinstance(data, list):
                scaled_data = []
                for element in data:
                    assert isinstance(element, pd.DataFrame)
                    scaled_data.append(element * scale_factor)
                all_data.append(scaled_data)
                lists = True
            else:
                all_data.append(data * scale_factor)
                lists = False
        if lists:
            return [combine(item, action='add') for item in all_data]
        return combine(all_data, action='add')

    def get_combined_basic_forces(
        self,
        recorder_name: str,
        combination: dict[str, float],
        live_load_reduction_factors: pd.Series | None = None,
    ) -> pd.DataFrame:
        """
        Get combined basic forces for a specific load combination.

        Raises:
          ValueError: If the recorder is not found for some load case.

        Returns:
          The results.
        """
        associated_load_cases = set(combination.keys())
        load_case_objects = {
            load_case: self.load_case_registry.find_load_case_by_name(load_case)
            for load_case in associated_load_cases
        }
        all_data = []
        for load_case_name, scale_factor in combination.items():
            load_case = load_case_objects[load_case_name]
            if recorder_name not in self.analysis_objects[load_case_name].recorders:
                msg = (
                    f'Recorder not found: {recorder_name} '
                    f'for load case {load_case}.'
                )
                raise ValueError(msg)
            data = self.analysis_objects[load_case_name].calculate_basic_forces(
                recorder_name,
                self.load_case_registry.model.components,
                ndm=NDM[self.load_case_registry.model.dimensionality],
                num_stations=12,
            )
            scaled_data = []
            for element in data:
                assert isinstance(element, pd.DataFrame)
                if (
                    '_live' in load_case_name
                    and live_load_reduction_factors is not None
                ):
                    element_names = element.columns.get_level_values(0)
                    reduction_factors = element_names.map(
                        live_load_reduction_factors.to_dict()
                    ).fillna(1.0)
                    scaled_data.append(element * scale_factor * reduction_factors)
                else:
                    scaled_data.append(element * scale_factor)
            all_data.append(scaled_data)
        return [combine([x[i] for x in all_data], action='add') for i in range(6)]



"""Defines :obj:`~osmg.model_objects.section.Section` objects."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import numpy as np

from osmg.core import common
from osmg.core.uid_object import UIDObject
from osmg.geometry import mesh
from osmg.geometry.mesh import Mesh, polygon_area

if TYPE_CHECKING:
    from shapely.geometry import Polygon as shapely_Polygon

    from osmg.core.common import numpy_array
    from osmg.model_objects.uniaxial_material import UniaxialMaterial
    from osmg.physical_material import PhysicalMaterial


@dataclass()
class Section(UIDObject):
    """
    Section object.

    The axes are defined in the same way as they are
    defined in OpenSees. The colors assigned to
    the axes for plotting follow the
    AutoCAD convention.

    .. code-block:: python

            y(green)
            ^         x(red)
            :       .
            :     .
            :   .
           ===
            | -------> z (blue)
           ===

    """

    name: str


@dataclass
class ElasticSection(Section):
    """
    Elastic Section Object.

    Attributes:
    ----------
      e_mod: Young's modulus.
      area: Cross-sectional area.
      i_y: Moment of inertia for strong-axis bending.
      i_x: Moment of inertia for weak-axis bending.
      g_mod: Shear modulus.
      j_mod: Torsional moment of inertia.
      sec_w: Weight per unit length.
      outside_shape: Mesh defining the outside shape of the section.
      snap_points: Dictionary containing coordinates of `snap_points`
        used by component-generating methods to position components
        relative to existing ones. See
        :func:`~osmg.creators.component.beam_placement_lookup` for example.
      properties: Dictionary containing section properties.

    """

    e_mod: float
    area: float
    i_y: float
    i_x: float
    g_mod: float
    j_mod: float
    sec_w: float
    outside_shape: Mesh | None = field(default=None, repr=False)
    snap_points: dict[str, numpy_array] | None = field(default=None, repr=False)
    properties: dict[str, Any] | None = field(default=None, repr=False)

    def weight_per_length(self) -> float:
        """
        Weight per unit length.

        Returns the weight per length of a section.
        For steel W sections, it adds 15% for misc. steel and
        connections.

        Returns:
          The weight per unit length.
        """
        if self.name[0] == 'W':
            res = self.sec_w * 1.15  # misc steel and connections
        else:
            res = self.sec_w
        return res

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'ElasticSection object\n'
        res += f'name: {self.name}\n'
        res += f'uid: {self.uid}\n'
        res += 'Properties:'
        res += f'  E: {self.e_mod}\n'
        res += f'  A: {self.area}\n'
        res += f'  Iy: {self.i_y}\n'
        res += f'  Ix: {self.i_x}\n'
        res += f'  G: {self.g_mod}\n'
        res += f'  J: {self.j_mod}\n'
        res += f'  W: {self.sec_w}\n'
        if self.outside_shape:
            res += 'outside_shape: specified\n'
        else:
            res += 'outside_shape: None\n'
        if self.snap_points:
            res += 'snap_points: specified\n'
        else:
            res += 'snap_points: None\n'
        return res


@dataclass(repr=False)
class SectionComponent:
    """
    Part of a section object, having a single material.

    Arguments:
      outside_shape: Mesh defining the outside shape
      ops_material: OpenSees material
      physical_material: Physical material
      parent_section: Parent section.
        The parent section is assigned automatically by their
        parent section iteslf, at its creation time.

    """

    outside_shape: Mesh
    holes: list[Mesh]
    ops_material: UniaxialMaterial
    physical_material: PhysicalMaterial
    parent_section: FiberSection | None = field(default=None)

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'SectionComponent object\n'
        if self.outside_shape:
            res += 'outside_shape: specified\n'
        else:
            res += 'outside_shape: None\n'
        if self.holes:
            res += 'holes: exist\n'
        else:
            res += 'holes: no holes\n'
        res += f'ops_material: {self.ops_material.name}\n'
        res += f'physical_material: {self.physical_material.name}\n'
        return res

    def cut_into_tiny_little_pieces(self) -> list[shapely_Polygon]:
        """
        Obtain data used to define fibers in OpenSees.

        Returns:
          The data.
        """
        # if we have an AISC HSS section, we need to discretize in a
        # certain way
        assert self.parent_section
        sec_name = self.parent_section.name
        rectangular_sections = 3
        circular_sections = 2
        if 'HSS' in sec_name and len(sec_name.split('X')) == rectangular_sections:
            # rectangular HSS section!
            pieces = mesh.subdivide_hss_rect(
                self.parent_section.properties['Ht'],
                self.parent_section.properties['B'],
                self.parent_section.properties['tdes'],
            )
        elif 'HSS' in sec_name and len(sec_name.split('X')) == circular_sections:
            # circular HSS section!
            pieces = mesh.subdivide_hss_circ(
                self.parent_section.properties['OD'],
                self.parent_section.properties['tdes'],
            )

        else:
            # fallback: use the default rectangular mesh chopper
            pieces = mesh.subdivide_polygon(
                outside=self.outside_shape,
                holes=self.holes,
                n_x=self.parent_section.n_x,
                n_y=self.parent_section.n_y,
            )

        return pieces


@dataclass(repr=False)
class FiberSection(Section):
    """
    Fiber section object.

    Can consist of multiple materials.
    The primary part of the component must have the key "main".

    """

    outside_shape: Mesh
    section_parts: dict[str, SectionComponent]
    j_mod: float
    snap_points: dict[str, numpy_array]
    properties: dict[str, Any]
    n_x: int
    n_y: int

    def __post_init__(self) -> None:
        """Post-initialization."""
        for part in self.section_parts:
            self.section_parts[part].parent_section = self

    def __repr__(self) -> str:
        """
        Get string representation.

        Returns:
          The string representation of the object.
        """
        res = ''
        res += 'FiberSection object\n'
        for part in self.section_parts:
            res += part.__repr__()
        if self.snap_points:
            res += 'snap_points: specified\n'
        else:
            res += 'snap_points: None\n'
        res += f'n_x: {self.n_x}, n_y: {self.n_y}\n'
        return res

    def ops_args(self) -> list[object]:
        """
        Obtain the OpenSees arguments.

        Returns:
          The OpenSees arguments.
        """
        return [
            'Fiber',
            self.uid,
            '-GJ',
            self.j_mod * self.section_parts['main'].physical_material.g_mod,
        ]

    def weight_per_length(self) -> float:
        """
        Weight per unit length.

        Returns the weight per length of a section.
        For steel W sections, it adds 15% for misc. steel and connections.

        Returns:
          The weight per unit length.
        """
        if self.name[0] == 'W':
            mult = 1.15  # misc steel and connections
        else:
            mult = 1.00
        res = 0.00
        for part in self.section_parts.values():
            coordinates: numpy_array = np.array(
                [h.vertex.coordinates for h in part.outside_shape.halfedges]
            )
            area = polygon_area(coordinates)
            for hole in part.holes:
                hole_coordinates: numpy_array = np.array(
                    [h.vertex.coordinates for h in hole.halfedges]
                )
                area -= polygon_area(hole_coordinates)
            density = part.physical_material.density
            # TODO(JVM): units
            res += area * density * common.G_CONST_IMPERIAL
        return res * mult
